@@ -2,12 +2,12 @@
 """
 Module containing functions and objects related to manipulating collections of atoms
 """
-import copy
 import logging
 import random
 from typing import Optional, Type, Tuple, Union, List
 
 import numpy as np
+import numpy.ma as ma
 from scipy.spatial.distance import pdist
 
 from . import functions
@@ -32,13 +32,6 @@ class AtomsCollection(functions.PlainState):
         the x, y, z positions and the atomic specie but this may change."""
         return 4
 
-    @staticmethod
-    def contains_num_atoms(state: functions.StateLike) -> int:
-        length = len(state)
-        if (length % AtomsCollection.num_atomic_properties()) != 0:
-            raise ValueError('AtomsCollection state vectors must be a multiple of 4')
-        return int(length / AtomsCollection.num_atomic_properties())
-
     @staticmethod
     def total_length(num_atoms: int) -> int:
         return AtomsCollection.num_atomic_properties() * num_atoms
@@ -208,7 +201,13 @@ class AtomsCollectionBuilder(functions.Function):
         def output_length(self, _in_state: AtomsCollection) -> int:
             return self._builder.input_length
 
-        def evaluate(self, state: AtomsCollection, *, get_jacobian=False) -> np.ndarray:
+        def evaluate(
+            # pylint: disable=unused-argument
+            self,
+            state: AtomsCollection,
+            *,
+            get_jacobian=False
+        ) -> np.ndarray:
             output_length = self.output_length(state)
             if self.output_length(state) != output_length:
                 raise ValueError(f'Expected input of length {output_length} but got {state.vector.size}')
@@ -218,7 +217,7 @@ class AtomsCollectionBuilder(functions.Function):
 
 
 class FeatureMapper(functions.Function):
-    """Map a collection of atoms onto a generic feature type such as Delta functions or Gaussians"""
+    """Map a collection of atoms onto a generic feature type such as delta functions or Gaussians"""
     input_type = AtomsCollection
     output_type = functions.Features
     supports_jacobian = True
@@ -244,34 +243,47 @@ class FeatureMapper(functions.Function):
         desc = [self._feature_type.__name__]
         if self._feature_kwargs:
             desc.append(f'({self._feature_type})')
-        desc.append(f' (species=>{self._map_species_to})')
-        return ''.join(desc)
+        desc.append(f'(species -> {self._map_species_to})')
+        mapping = ''.join(desc)
+        return f'FeatureMapper({mapping})'
 
     @property
     def inverse(self):
         return self._inverse
 
     @property
-    def map_species_to(self) -> Optional[str]:
+    def map_species_to(self) -> Optional[int]:
         return self._map_species_to
 
     @staticmethod
     def _get_species_map_idx(feature_type: Type[functions.Feature], map_to: Union[int, str]) -> Optional[int]:
-        if isinstance(map_to, int):
-            return map_to
-        if isinstance(map_to, str):
-            return getattr(feature_type, map_to)
         if map_to is None:
             return None
 
-        raise TypeError(map_to)
+        map_idx = None
+        if isinstance(map_to, int):
+            map_idx = map_to
+        elif isinstance(map_to, str):
+            map_idx = getattr(feature_type, map_to)
+        else:
+            raise TypeError(map_to)
+
+        if not isinstance(map_idx, int):
+            raise TypeError(f"Expected '{map_to}' to correspond to an integer index, but got '{map_idx}'")
+
+        return map_idx
 
     def output_length(self, in_state: AtomsCollection) -> int:
         return in_state.num_atoms * self._feature_type.LENGTH
 
-    def evaluate(self, atoms: AtomsCollection, *, get_jacobian=False) -> functions.Features:  # pylint: disable=arguments-differ
+    def evaluate(
+        # pylint: disable=arguments-differ
+        self,
+        atoms: AtomsCollection,
+        *,
+        get_jacobian=False
+    ) -> functions.Features:
         features = functions.Features()
-
         jac = np.zeros((self.output_length(atoms), len(atoms))) if get_jacobian else None
 
         # Create the features one by one and add them to the features vector
@@ -288,9 +300,10 @@ class FeatureMapper(functions.Function):
                 # where we optionally map the species to the index s
                 pos_idx = atoms.linear_pos_idx(atom_idx).start
                 for i in range(0, 3):
-                    jac[idx + i, pos_idx + i] = 1.
+                    jac[idx + i, pos_idx + i] = 1.  # pylint: disable=unsupported-assignment-operation
 
                 if self._map_species_to:
+                    # pylint: disable=unsupported-assignment-operation
                     jac[idx + self._map_species_to, atoms.linear_number_idx(atom_idx)] = 1.
 
             features.add(feature)
@@ -314,13 +327,18 @@ class FeatureMapper(functions.Function):
         def output_length(features: functions.Features) -> int:
             return AtomsCollection.total_length(len(features.features))
 
-        def evaluate(self, features: functions.Features, get_jacobian=False):  # pylint: disable=arguments-differ
+        def evaluate(
+            # pylint: disable=unused-argument, arguments-differ
+            self,
+            features: functions.Features,
+            get_jacobian=False
+        ):
             atoms_collection = AtomsCollection(len(features.features))
 
             for idx, feature in enumerate(features.features):
                 atoms_collection.positions[idx] = feature.pos
                 if self._mapper.map_species_to:
-                    atoms_collection.numbers[idx] = getattr(feature, self._mapper.map_species_to)
+                    atoms_collection.numbers[idx] = feature.vector[self._mapper.map_species_to]
 
             return atoms_collection
 
@@ -376,7 +394,13 @@ class CentreAtomsCollection(functions.Function):
     def output_length(in_state: AtomsCollection) -> int:
         return len(in_state)
 
-    def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False) -> AtomsCollection:  # pylint: disable=arguments-differ
+    def evaluate(
+        # pylint: disable=unused-argument, arguments-differ
+        self,
+        in_atoms: AtomsCollection,
+        *,
+        get_jacobian=False
+    ) -> AtomsCollection:
         out_atoms = AtomsCollection(in_atoms.num_atoms)
 
         centre = np.sum(in_atoms.positions) / in_atoms.num_atoms
@@ -390,7 +414,7 @@ class MapNumbers(functions.Function):
     """Map the given set of atom numbers onto a continuous range"""
     input_type = AtomsCollection
     output_type = AtomsCollection
-    supports_jacobian = False
+    supports_jacobian = True
 
     def __init__(self, species: set, map_to: Union[float, Tuple[float, float]] = (1.0, 6.0)):
         super().__init__()
@@ -421,7 +445,12 @@ class MapNumbers(functions.Function):
     def output_length(in_state: AtomsCollection) -> int:
         return len(in_state)
 
-    def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False) -> AtomsCollection:  # pylint: disable=arguments-differ
+    def evaluate(
+        # pylint: disable=unused-argument, arguments-differ
+        self,
+        in_atoms: AtomsCollection,
+        get_jacobian=False
+    ) -> Union[AtomsCollection, Tuple[AtomsCollection, np.ndarray]]:
         out_atoms = in_atoms.copy()
 
         # Now adjust the numbers
@@ -435,6 +464,16 @@ class MapNumbers(functions.Function):
                     new_numbers = (num_idx / len(self._numbers)) * self._range_size + self._mapped_range[0]
                     out_atoms.numbers[idx] = new_numbers + self._half_bin
 
+        if get_jacobian:
+            # Create Jacobian that is identity apart from the species.
+            # This part needs to be masked of as this mapping is not a continuous function
+            natoms = in_atoms.num_atoms
+            jac = ma.array(np.zeros((len(out_atoms), len(in_atoms))), mask=False)
+            jac[:3 * natoms, :3 * natoms] = np.eye(natoms * 3)
+            jac.mask[3 * natoms:, 3 * natoms:] = True
+
+            return out_atoms, jac
+
         return out_atoms
 
     class Inverse(functions.Function):
@@ -449,13 +488,22 @@ class MapNumbers(functions.Function):
         def output_length(in_state: AtomsCollection) -> int:
             return len(in_state)
 
-        def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False) -> AtomsCollection:  # pylint: disable=arguments-differ
+        def evaluate(
+            # pylint: disable=unused-argument, arguments-differ
+            self,
+            in_atoms: AtomsCollection,
+            *,
+            get_jacobian=False
+        ) -> AtomsCollection:  # pylint: disable=arguments-differ
             out_atoms = in_atoms.copy()
 
             # Now adjust the numbers
             for idx, num in enumerate(out_atoms.numbers):
                 if self._mapped_range[0] <= num <= self._mapped_range[1]:
-                    rescaled = (num - self._mapped_range[0]) / self._range_size * len(self._numbers)
+                    if self._range_size == 0.:
+                        rescaled = self._mapped_range[0]
+                    else:
+                        rescaled = (num - self._mapped_range[0]) / self._range_size * len(self._numbers)
                     out_atoms.numbers[idx] = self._numbers[int(rescaled)]
                 else:
                     _LOGGER.warning(
@@ -507,8 +555,8 @@ class ApplyCutoff(functions.Function):
         return out_atoms
 
 
-def random_atom_collection_in_sphere(num: int, radius=1., centre=True, numbers=1.) -> AtomsCollection:
-    pts = generate.random_points_in_sphere(num, radius=radius, centre=centre)
+def random_atom_collection_in_sphere(num: int, radius=1., centre=True, numbers=1., minsep=None) -> AtomsCollection:
+    pts = generate.random_points_in_sphere(num, radius=radius, centre=centre, minsep=minsep)
     atoms = AtomsCollection(num, positions=pts)
     if isinstance(numbers, tuple):
         atoms.numbers[:] = random.choices(numbers, k=num)
@@ -521,11 +569,22 @@ def random_atom_collection_in_sphere(num: int, radius=1., centre=True, numbers=1
 class SeparationForce(functions.Function):
     """A separation force between atoms that come within the cutoff.  This can be used as a penalty in a loss function
     to """
+    supports_jacobian = True
+    output_type = float
 
-    def __init__(self, force_constant: float, cutoff: float):
+    def __init__(self, epsilon: float = 1, sigma=1.0, cutoff=1., power: int = 12):
         super().__init__()
-        self._force_constant = force_constant
+        self._epsilon = epsilon
+        self._sigma = sigma
         self._cutoff = cutoff
+        self._power = power
+
+        if cutoff is None:
+            self._c1 = 0
+            self._c2 = 0
+        else:
+            self._c1 = self._power * self._sigma**self._power / cutoff**(self._power + 1)
+            self._c2 = -(self._sigma / cutoff)**self._power - self._c1 * cutoff
 
     def evaluate(self, atoms: AtomsCollection, *, get_jacobian=False):  # pylint: disable=arguments-differ
         distances = pdist(atoms.positions)
@@ -539,7 +598,7 @@ class SeparationForce(functions.Function):
             # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.pdist.html#scipy.spatial.distance.pdist
             idx = lambda i, j: n * i + j - ((i + 2) * (i + 1)) // 2
             for i in range(n):
-                for j in range(i, n):
+                for j in range(i + 1, n):
                     dr = positions[j] - positions[i]  # pylint: disable=invalid-name
                     dist = distances[idx(i, j)]
                     force = self.force(dist) * dr / dist
@@ -547,38 +606,39 @@ class SeparationForce(functions.Function):
                     jacobian[0, atoms.linear_pos_idx(i)] += force
                     jacobian[0, atoms.linear_pos_idx(j)] -= force
 
+            # return 0., np.zeros(jacobian.shape)
             return total_energy, jacobian
 
+        # return 0.
         return total_energy
 
+    @property
+    def cutoff(self) -> float:
+        """The distance at which this force and energy goes to zero"""
+        return self._cutoff
+
     def energy(  # pylint: disable=invalid-name
         self, r: float
     ) -> float:
         """
-        Energy is
-            k (r^2 / 2 - r r0) + r0
-        where k is force constant and r0 is cutoff
-
         :param r: separation between atoms
         :return: the corresponding energy
         """
-        if r > self._cutoff:
+        if self._cutoff is not None and r > self.cutoff:
             return 0.
 
-        return self._force_constant * r * (0.5 * r - self._cutoff) + self._cutoff
+        energy = 4 * self._epsilon * ((self._sigma / r)**self._power + self._c1 * r + self._c2)
+        return energy
 
     def force(  # pylint: disable=invalid-name
         self, r: float
     ):
         """
-        Force is
-            -k (r - r0)
-        where k is force constant and r0 is cutoff
-
         :param r: separation between atoms
         :return: the corresponding force magnitude
         """
-        if r > self._cutoff:
+        if self._cutoff is not None and r > self.cutoff:
             return 0.
 
-        return -self._force_constant * (r - self._cutoff)
+        force = 4 * self._epsilon * (self._power * self._sigma**self._power / r**(self._power + 1) - self._c1)
+        return force
