@@ -8,8 +8,12 @@ from typing import List, Union, Tuple, Callable, Any, Iterable, Type, Optional
 import numpy as np
 import scipy
 
+from . import mathutil
+
 _LOGGER = logging.getLogger(__name__)
 
+SQRT_TWO_PI = (2 * np.pi)**0.5
+
 
 class State(metaclass=abc.ABCMeta):
     """A state vector representing an input to a function"""
@@ -25,7 +29,7 @@ class State(metaclass=abc.ABCMeta):
 
     @property
     def size(self):
-        return self.array.size
+        return self.vector.size
 
     @property
     def array(self) -> np.array:
@@ -61,6 +65,12 @@ class PlainState(State):
     def vector(self) -> np.array:
         return self._array
 
+    @vector.setter
+    def vector(self, value: np.array):
+        if not value.shape == self._array.shape:
+            raise ValueError(f"Shape mismatch, expected '{self._array.shape}', got '{value.shape}'")
+        self._array = value
+
     @property
     def array(self) -> np.array:
         """Get the array corresponding to this state"""
@@ -103,9 +113,9 @@ class Feature(State):
 
     @vector.setter
     def vector(self, value: np.array):
-        if not value.size == self._vector.size:
-            raise ValueError("Size mismatch, expected '{}', got '{}'".format(self._vector.size, value.size))
-        self._vector[:] = value[:]
+        if not value.shape == self._vector.shape:
+            raise ValueError(f"Shape mismatch, expected '{self._vector.shape}', got '{value.shape}'")
+        self._vector = value
 
     def __add__(self, other: 'Feature') -> 'Features':
         if not isinstance(other, Feature):
@@ -114,7 +124,7 @@ class Feature(State):
 
 
 class Features(State):
-    """An collection of features"""
+    """A collection of features"""
 
     def __init__(self, *feature: Feature):
         self._vector = np.empty(0)
@@ -122,6 +132,9 @@ class Features(State):
         for entry in feature:
             self.add(entry)
 
+    def __getitem__(self, item):
+        return self._features[item]
+
     @property
     def features(self) -> List[Feature]:
         """Access the list of features"""
@@ -135,9 +148,20 @@ class Features(State):
     def array(self):
         return self._vector
 
-    @array.setter
-    def array(self, arr):
-        self._vector[:] = arr
+    @vector.setter
+    def vector(self, vec):
+        # Reassign all the array
+        if self._vector.shape != vec.shape:
+            raise ValueError(
+                f'The passed vector ({vec.shape}) has a different shape from the features '
+                f'vector ({self._vector.shape})'
+            )
+        self._vector = vec
+
+        idx = 0
+        for entry in self._features:
+            entry.vector = self._vector[idx:idx + len(entry)]
+            idx += entry.length
 
     def feature_range(self, feature: Feature) -> Optional[range]:
         """Given a feature in this set of features, get the corresponding slice of indices in the feature vector"""
@@ -177,7 +201,7 @@ class WeightedDelta(Feature):
     LENGTH = 4
 
     def __init__(self, pos: np.array, weight=1.0):
-        super().__init__(dtype=pos.dtype)
+        super().__init__(dtype=np.promote_types(pos.dtype, type(weight)))
         self.pos = pos
         self.weight = weight
 
@@ -245,6 +269,36 @@ class WeightedGaussian(Feature):
         self.vector[self.WEIGHT] = value
 
 
+class ListState(State):
+
+    def __init__(self, *entries):
+        self._list = list(*entries)
+
+    def append(self, state: StateLike):
+        if not isinstance(state, StateLike):
+            raise TypeError(f"Expected state like object, got '{type(state)}'")
+
+        self._list.append(state)
+
+    def __len__(self):
+        return self._list.__len__()
+
+    def __iter__(self):
+        return self._list.__iter__()
+
+    @property
+    def vector(self) -> np.array:
+        return np.hstack(tuple(map(get_bare, self._list)))
+
+    @vector.setter
+    def vector(self, new_val: np.ndarray):
+        i = 0
+        for entry in self._list:
+            length = len(entry)
+            set_bare(entry, new_val[i:i + length])
+            i += length
+
+
 # region Functions
 
 
@@ -270,6 +324,7 @@ class Function(metaclass=abc.ABCMeta):
         self._callbacks.remove(func)
 
     def __call__(self, state: State, jacobian=False):
+        # pylint: disable=too-many-branches
         name = self.__class__.__name__
 
         if self.input_type is not None:
@@ -376,6 +431,7 @@ class FromVectorBuilder(Function):
 
 class Identity(Function):
     """Function that just returns what it is passed"""
+    supports_jacobian = True
 
     def evaluate(self, state: StateLike, *, get_jacobian=False):
         """Evaluate the function with the passed input"""
@@ -486,7 +542,7 @@ class Chain(Function):
                     previous_jacobian = current_jacobian
                 else:
                     # Apply chain rule and propagate jacobians
-                    previous_jacobian = np.matmul(current_jacobian, previous_jacobian)
+                    previous_jacobian = matmul(current_jacobian, previous_jacobian)
 
             # Make this output the input to the next function
             current_in = current_out
@@ -497,26 +553,31 @@ class Chain(Function):
         current_out = function(current_in, get_jacobian)
         if get_jacobian:
             current_out, current_jacobian = current_out
-            return current_out, np.matmul(current_jacobian, previous_jacobian)
+
+            return current_out, matmul(current_jacobian, previous_jacobian)
 
         return current_out
 
 
+def matmul(mat1, mat2):
+    try:
+        return mat1 @ mat2
+    except ValueError:
+        # Use einsum here as matrix multiply fro masked arrays doesn't work yet, see:
+        # https://github.com/numpy/numpy/issues/14992
+        # https://stackoverflow.com/questions/63218951/masked-matrix-multiplication
+        return np.einsum('ij,jk', mat1, mat2)
+
+
 class Residuals(Function):
     """Given some measurements (m) calculate the residuals from some data (d) as f = d - m """
     supports_jacobian = True
 
-    def __init__(self, data: np.array):
+    def __init__(self, data: StateLike):
         super().__init__()
         self._data = get_bare(data)
 
-    def output_length(self, in_state: State) -> int:  # pylint: disable=unused-argument
-        return len(self._data)
-
-    def empty_jacobian(self, in_state: State) -> np.array:
-        return np.empty((self.output_length(in_state), len(in_state)), dtype=complex)
-
-    def evaluate(self, state: State, *, get_jacobian=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
+    def evaluate(self, state: StateLike, *, get_jacobian=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
         out_vector = get_bare(state) - self._data
         if get_jacobian:
             return out_vector, np.identity(len(state), dtype=out_vector.dtype)
@@ -543,11 +604,14 @@ class Native(Function):
 
 
 class CosineCutoff(Function):
+    # pylint: disable=anomalous-backslash-in-string
     """A function that takes a set of Features and scales their weights by applying a cosine cutoff in the form
 
-    w' = w (cos(\pi * dr / r_\text{cut}) + 1.) / 2
+        w' = w (cos(\pi * dr / r_\text{cut}) + 1.) / 2
     """
-    supports_jacobian = False
+    input_type = (Features,)
+    output_type = (Features,)
+    supports_jacobian = True
 
     def __init__(self, cutoff: float):
         super().__init__()
@@ -623,6 +687,81 @@ class CosineCutoff(Function):
         return weight * (sp.cos(sp.pi * dr / rcut) + 1) / 2
 
 
+class SphericalToCart(Function):
+
+    def evaluate(
+        # pylint: disable=arguments-differ
+        self,
+        spherical: np.ndarray,
+        *,
+        get_jacobian=False
+    ):
+        cart = mathutil.sph2cart(spherical)
+
+        if get_jacobian:
+            pass
+
+        return cart
+
+
+class MeanSquaredError(Function):
+    """Calcualte the MSE relative to some target value(s)"""
+    supports_jacobian = True
+    output_type = float, complex
+
+    def __init__(self, target: StateLike):
+        super().__init__()
+        self._target = get_bare(target)
+
+    def evaluate(
+        # pylint: disable=arguments-differ
+        self,
+        values: StateLike,
+        *,
+        get_jacobian=False
+    ):
+        # pylint: disable=invalid-name
+        diff = get_bare(values) - self._target
+        n = len(diff)
+        mse = np.sum(diff**2) / n
+
+        if get_jacobian:
+            return mse, (2 * diff / n).reshape(1, n)
+
+        return mse
+
+
+class Map(Function):
+    """Function that passes the same input to multiple functions and returns an ordered array of their outputs"""
+    supports_jacobian = True
+    output_type = ListState
+
+    def __init__(self, *fn):
+        super().__init__()
+        self._fns = fn
+
+    def evaluate(self, state, *, get_jacobian=False):
+        if get_jacobian:
+            output, jacs = zip(*tuple(fn(state, jacobian=True) for fn in self._fns))
+            jac = np.concatenate(jacs)
+
+            return ListState(output), jac
+
+        return ListState(tuple(fn(state) for fn in self._fns))
+
+
+class HStack(Function):
+    supports_jacobian = True
+    output_type = np.ndarray
+
+    def evaluate(self, state, *, get_jacobian=False):
+        outputs = np.hstack(tuple(map(get_bare, state)))
+        if get_jacobian:
+            return outputs, np.eye(len(outputs))
+
+        return outputs
+
+
 # endregion
 
 
@@ -636,6 +775,13 @@ def get_bare(state: Union[np.ndarray, State]) -> Union[np.array, numbers.Number]
     return state
 
 
+def set_bare(state: Union[np.ndarray, State], new_val: np.ndarray):
+    if isinstance(state, np.ndarray):
+        state[:] = new_val
+    else:
+        state.vector = new_val
+
+
 def copy_to(
     dest: np.ndarray,
     indices: np.ndarray,
