@@ -47,9 +47,9 @@ class GeometricMoments(base_moments.Moments):
         # Just trying this for now, let's see if it's possible to not force self._moments to be a numpy array
         max_order = moments.shape[0] - 1
         # Mask everything and then unmask the ones we want
-        mask = np.ones(moments.shape, dtype=np.int)
-        idx_array = np.empty(moments.shape, dtype=np.int)
-        for i, indices in enumerate(iter_indices(max_order)):
+        mask = np.ones(moments.shape, dtype=int)
+        idx_array = np.empty(moments.shape, dtype=int)
+        for i, indices in enumerate(get_indices(max_order)):
             mask[indices] = 0
             idx_array[indices] = i
 
@@ -79,10 +79,24 @@ class GeometricMoments(base_moments.Moments):
     def vector(self) -> np.array:
         return self._moments.compressed()
 
+    @vector.setter
+    def vector(self, value: np.ndarray):
+        if value.dtype == self._moments.dtype:
+            self._moments[~self._moments.mask] = value
+        else:
+            # Convert the dtype first and then set values
+            new_moments = self._moments.astype(value.dtype)
+            new_moments[~new_moments.mask] = value
+            self._moments = new_moments
+
     @property
     def moments(self) -> np.array:
         return self._moments
 
+    @property
+    def array(self) -> np.ndarray:
+        return self._moments
+
     def to_matrix(self) -> np.array:
         return self._moments
 
@@ -100,14 +114,16 @@ class GeometricMoments(base_moments.Moments):
         """Reconstruct the value at x from the moments
 
         :param x: the point to get the value at
-        :param max_order: the maximum order to go up to (defaults to the maximum order of these
-            moments)
+        :param max_order: the maximum order to go up to (defaults to the maximum order of these moments)
         """
         shape = self._moments.shape
         value = 0.
         for p in range(shape[0]):
             for q in range(shape[1]):
                 for r in range(shape[2]):
+                    if (p + q + r) > max_order:
+                        continue
+
                     value += self._moments[p, q, r] * (x**(p, q, r)).prod(axis=-1)
 
         return value
@@ -117,7 +133,11 @@ class GeometricMoments(base_moments.Moments):
         geom_moms.fill(None)
         return GeometricMoments(geom_moms)
 
-    def get_builder(self, mask=None) -> Optional[functions.Function]:
+    def get_builder(
+        # pylint: disable=unused-argument
+        self,
+        mask=None
+    ) -> Optional[functions.Function]:
         return functions.FromVectorBuilder(GeometricMoments, kwargs=dict(max_order=self.max_order))
 
 
@@ -129,6 +149,8 @@ def linear_index(max_order: int, index: base_moments.Index) -> int:
         if indices == index:
             return i
 
+    raise RuntimeError('Should never reach here')
+
 
 def iter_indices(max_order: int) -> Iterator[base_moments.Index]:
     for i in utils.inclusive(max_order):
@@ -137,6 +159,11 @@ def iter_indices(max_order: int) -> Iterator[base_moments.Index]:
                 yield GeometricIndex(i, j, k)
 
 
+@functools.lru_cache(maxsize=None)
+def get_indices(max_order: int):
+    return tuple(iter_indices(max_order))
+
+
 class GeometricMomentsCalculator(base_moments.MomentsCalculator):
     """Function that calculates geometric moments up to some maximum order"""
 
@@ -184,7 +211,7 @@ def _(delta: functions.WeightedDelta, max_order: int, get_jacobian: bool, dtype)
     moments = delta.weight * utils.outer_product(moms[:, 0], moms[:, 1], moms[:, 2])
 
     if get_jacobian:
-        indices = tuple(iter_indices(max_order))
+        indices = get_indices(max_order)
         jacobian = np.zeros((len(indices), delta.LENGTH), dtype=dtype)
 
         for i, (p, q, r) in enumerate(indices):
@@ -204,22 +231,25 @@ def _(gaussian: functions.WeightedGaussian, max_order: int, get_jacobian: bool,
     O = max_order
 
     partial_moments = np.empty((O + 1, 3), dtype=dtype)
-    partial_moments[0, :] = 1.0  # 0^th order, mass is multiplied in at end
+    partial_moments[0, :] = 1  # 0^th order, mass is multiplied in at end
 
     for order in utils.inclusive(0, O):
-        partial_moments[order, :] = np.array(gaussian_moment[order](gaussian.pos, gaussian.sigma))
+        partial_moments[order, :] = np.array(gaussian_moment[order](gaussian.pos, gaussian.sigma), dtype=dtype)
 
     moments = utils.outer_product(partial_moments[:, 0], partial_moments[:, 1], partial_moments[:, 2])
 
     if get_jacobian:
-        dg = np.empty((
-            O + 1,  # Order
-            2,  # d/dmu, d/dsigma
-            3,  # x, y, z
-        ))
+        dg = np.empty(
+            (
+                O + 1,  # Order
+                2,  # d/dmu, d/dsigma
+                3,  # x, y, z
+            ),
+            dtype=dtype
+        )
 
         for order in utils.inclusive(max_order):
-            dg[order, :, :] = np.array(gaussian_moment_derivatives[order](gaussian.pos, gaussian.sigma))
+            dg[order, :, :] = np.array(gaussian_moment_derivatives[order](gaussian.pos, gaussian.sigma), dtype=dtype)
 
         # Let's get the indices and create a Jacobian matrix
         indices = tuple(iter_indices(max_order))
@@ -238,8 +268,8 @@ def _(gaussian: functions.WeightedGaussian, max_order: int, get_jacobian: bool,
                     partial_moments[p, 0] * partial_moments[q, 1] * dg[r, 1, 2]
             )
 
-        # Weights
-        jacobian[:, gaussian.WEIGHT] = moments[:, :, :]
+            # Weights
+            jacobian[i, gaussian.WEIGHT] = moments[p, q, r]
 
         return gaussian.weight * moments, jacobian
 
@@ -250,10 +280,7 @@ def _(gaussian: functions.WeightedGaussian, max_order: int, get_jacobian: bool,
 def _(environment: functions.Features, max_order: int, get_jacobian, dtype):
     """Calculate the geometric moments for a set of features"""
     O = max_order
-    idx = 0
-
-    moments = np.empty((O + 1, O + 1, O + 1), dtype=dtype)
-    moments.fill(0.)
+    moments = np.zeros((O + 1, O + 1, O + 1), dtype=dtype)
 
     jacobians = []
     for feature in environment.features:
@@ -265,10 +292,14 @@ def _(environment: functions.Features, max_order: int, get_jacobian, dtype):
 
         moments += moms
 
-        idx += len(feature)
-
     if get_jacobian:
-        return moments, np.concatenate(jacobians, axis=1)
+        if jacobians:
+            jac = np.concatenate(jacobians, axis=1)
+        else:
+            total = sum(1 for _ in iter_indices(O))
+            jac = np.zeros((total, 0))
+
+        return moments, jac
 
     return moments
 
@@ -277,7 +308,8 @@ def from_gaussians(
     max_order: int,
     positions: np.ndarray,
     sigmas: Union[numbers.Number, np.array] = 0.4,
-    weights: Union[numbers.Number, np.array] = 1.
+    weights: Union[numbers.Number, np.array] = 1.,
+    get_jacobian=False
 ) -> GeometricMoments:
     """Calculate the geometric moments for a collection of Gaussians at the given positions with
     the passed parameters.
@@ -297,7 +329,7 @@ def from_gaussians(
     for pos, sigma, weight in zip(positions, sigmas, weights):
         features.add(functions.WeightedGaussian(pos, sigma, weight))
 
-    return GeometricMomentsCalculator(max_order)(features)
+    return GeometricMomentsCalculator(max_order)(features, jacobian=get_jacobian)
 
 
 def from_deltas(
@@ -359,11 +391,11 @@ def from_deltas_analytic(max_order: int, num_particles: int, pos_symbols=None, w
 
 
 def gaussian_moment_0(mu: float, sigma: float = 1.) -> float:  # pylint: disable=unused-argument
-    return 1.
+    return 1
 
 
 def gaussian_moment_0_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:  # pylint: disable=unused-argument
-    return 0., 0.
+    return 0, 0
 
 
 def gaussian_moment_1(mu: float, sigma: float = 1.) -> float:  # pylint: disable=unused-argument
@@ -371,7 +403,7 @@ def gaussian_moment_1(mu: float, sigma: float = 1.) -> float:  # pylint: disable
 
 
 def gaussian_moment_1_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:  # pylint: disable=unused-argument
-    return 1., 0.
+    return 1, 0
 
 
 def gaussian_moment_2(mu: float, sigma: float = 1.) -> float:
@@ -563,6 +595,7 @@ def gaussian_moment_16(mu: float, sigma: float = 1.) -> float:
            2027025 * sigma ** 16
 
 
+# pylint: disable=pointless-string-statement
 """Get the nt^h moment of a n-dim Gaussian (or normal distribution) centred at `mu`
 with a standard deviation of `sigma`.
 
