@@ -1,54 +1,25 @@
 # -*- coding: utf-8 -*-
 import collections
 import math
-from typing import Union, Callable, Sequence, Optional, List
+from typing import Union, Callable, Sequence, Optional, List, Tuple, Iterator
 
 import ase
+
+try:
+    import functorch
+except ImportError:
+    functorch = None
 import torch
 import torch.nn.functional
 from torch.autograd import functional
 
-from milad.play import asetools
-from milad import fingerprinting
-from milad import utils
+from milad import dat
 
-__all__ = 'NeuralNetwork', 'create_fingerprint_set'
+__all__ = ('NeuralNetwork',)
 
 # pylint: disable=no-member, not-callable
 
 
-def create_fingerprint_set(
-    descriptor: fingerprinting.MomentInvariantsDescriptor,
-    systems: Sequence[ase.Atoms],
-    get_derivatives=False
-) -> utils.FingerprintSet:
-    # WARNING: The calculation of derivatives only takes into account positional degrees of freedom (not species) and
-    # makes assumptions about the shape of the derivatives tensor
-
-    fp_length = descriptor.fingerprint_len
-
-    fingerprints = utils.FingerprintSet(descriptor.fingerprint_len)
-    for system in systems:
-        fps = []
-        fp_derivatives = []
-        for env in asetools.extract_environments(system, cutoff=descriptor.cutoff):
-            milad_env = asetools.ase2milad(env)
-            if get_derivatives:
-                fingerprint, jacobian = descriptor(milad_env, jacobian=True)
-                # Now, let's deal with the derivatives extracting just the positional parts and summing over all
-                # neighbours keeping x, y, z separate
-                derivatives = jacobian[:, 3:3 * len(env)].reshape(fp_length, -1, 3).sum(1)
-
-                fps.append(fingerprint)
-                fp_derivatives.append(derivatives)
-            else:
-                fps.append(descriptor(milad_env))
-
-        fingerprints.add_system(system, fps, derivatives=fp_derivatives)
-
-    return fingerprints
-
-
 class Predictions:
     """Class representing predictions made by the the model"""
 
@@ -90,11 +61,16 @@ class Predictions:
         return self.total_energies / self.sizes
 
 
+SystemFittingData = collections.namedtuple(
+    'SystemFittingData', 'index num_atoms total_energy forces fingerprints derivatives'
+)
+
+
 class FittingData:
     """Data used during a fitting procedure"""
 
     @classmethod
-    def from_fingerprint_set(cls, fingerprint_set: utils.FingerprintSet, device=None, dtype=None, requires_grad=False):
+    def from_fingerprint_set(cls, fingerprint_set: dat.FingerprintSet, device=None, dtype=None, requires_grad=False):
         tensor_kwargs = dict(dtype=dtype, device=device, requires_grad=requires_grad)
         fingerprints = torch.cat(tuple(torch.tensor(fps, **tensor_kwargs) for fps in fingerprint_set.fingerprints))
 
@@ -109,15 +85,13 @@ class FittingData:
 
         derivatives = None
         if fingerprint_set.has_all_derivatives():
-            derivatives = torch.cat(
-                tuple(
-                    torch.tensor(derivatives, **tensor_kwargs) if derivatives is not None else None
-                    for derivatives in fingerprint_set.fingerprint_derivatives
-                )
+            derivatives = tuple(
+                torch.tensor(derivatives, **tensor_kwargs) if derivatives is not None else None
+                for derivatives in fingerprint_set.fingerprint_derivatives
             )
 
         return FittingData(
-            torch.tensor(fingerprint_set.sizes, dtype=torch.int, device=device),
+            torch.tensor(fingerprint_set.sizes, dtype=torch.int64, device=device),
             fingerprints,
             total_energies=torch.tensor(fingerprint_set.get_potential_energies(normalise=False), **tensor_kwargs),
             forces=forces,
@@ -130,11 +104,13 @@ class FittingData:
         fingerprints: torch.Tensor,
         total_energies: torch.Tensor,
         forces: Optional[torch.Tensor] = None,
-        derivatives: Optional[torch.Tensor] = None
+        derivatives: Tuple[Optional[torch.Tensor]] = None
     ):
         """
         Construct a set of fitting data.
 
+        :param sizes: a tensor containing the number of atoms in each system.  The length of the tensor should be equal
+            to that of total_energies
         :param fingerprints: a tuple containing a set of fingerprints for each atom in the structure
         :param total_energies: the corresponding energy for each structure
         :param forces: the (optional) forces for each atom in the structure
@@ -149,13 +125,16 @@ class FittingData:
                 f"The number of total energies ({len(total_energies)}) doesn't match the number of "
                 f'systems (sizes) ({len(sizes)}).'
             )
+        if derivatives and len(derivatives) != len(sizes):
+            raise ValueError("Number of system-wise derivatives doesn't match the number of sizes")
 
-        self._num_atoms = sizes
-        self._fingerprints = fingerprints
-        self._total_energies = total_energies
-        self._forces = forces
-        self._derivatives = derivatives
-        self._index = _create_index_from_sizes(self._num_atoms)
+        self._num_atoms = sizes  # Per-system
+        self._total_energies = total_energies  # Per-system
+        self._derivatives = derivatives  # Per system
+
+        self._fingerprints = fingerprints  # Per-environment
+        self._forces = forces  # Per-environment
+        self._index = _create_index_from_sizes(self._num_atoms)  # Per-environment
 
     def __getitem__(self, item) -> 'FittingData':
         if isinstance(item, slice):
@@ -165,7 +144,7 @@ class FittingData:
             total_energies = self._total_energies[item]
             fingerprints = self._fingerprints[atomic_slice]
             forces = self._forces[atomic_slice] if self._forces is not None else None
-            derivatives = self._derivatives[atomic_slice] if self._derivatives is not None else None
+            derivatives = self._derivatives[item] if self._derivatives is not None else None
 
             return FittingData(sizes, fingerprints, total_energies, forces, derivatives)
 
@@ -184,9 +163,28 @@ class FittingData:
         stop = self._num_atoms[:item.stop].sum().cpu().item()
         return slice(start, stop)
 
+    def iter_systemwise(self) -> Iterator[SystemFittingData]:
+        env_idx = 0
+        for system_idx, natoms in enumerate(self._num_atoms):
+            env_slice = slice(env_idx, env_idx + natoms)
+            yield SystemFittingData(
+                index=system_idx,
+                num_atoms=natoms,
+                total_energy=self._total_energies[system_idx],
+                derivatives=self._derivatives[system_idx],
+                fingerprints=self._fingerprints[env_slice],
+                forces=self._forces[env_slice],
+            )
+            env_idx += natoms
+
+    @property
+    def fingerprint_len(self) -> int:
+        """Get the length of the fingerprint vector"""
+        return self._fingerprints.shape[1]
+
     @property
     def fingerprints(self) -> torch.Tensor:
-        """Fingerprints tensors.  One entry per system."""
+        """Fingerprints tensors.  One entry per atomic environment."""
         return self._fingerprints
 
     @property
@@ -200,7 +198,10 @@ class FittingData:
         return self._forces
 
     @property
-    def derivatives(self) -> Optional[torch.Tensor]:
+    def derivatives(self) -> Tuple[Optional[torch.Tensor]]:
+        """Get the derivatives for each environment in the structure.  This returns a tuple with one entry per system.
+
+        Each entry is a tensor in the format [central atom idx, neighbour idx, 3 (x, y, z), fingerprint length]"""
         return self._derivatives
 
     @property
@@ -210,8 +211,8 @@ class FittingData:
 
     @property
     def index(self) -> torch.Tensor:
-        """Get an tensor of structure indices that each local environment datum belongs to
-        (e.g. force, derivatices, etc)"""
+        """Get a tensor of structure indices that each local environment datum belongs to
+        (e.g. force, derivatives, etc)"""
         return self._index
 
     def get_normalised_energies(self) -> torch.Tensor:
@@ -220,6 +221,9 @@ class FittingData:
     def batch_split(self, batchsize: int) -> List['FittingData']:
         total_samples = len(self)
         num_batches = int(math.ceil(total_samples / batchsize))
+        if num_batches == 1:
+            return [self]
+
         batches = []
         for batch_num in range(num_batches):
             start_idx = batch_num * batchsize
@@ -292,13 +296,12 @@ class Range:
             self._range[1] = self.max if self.max > vals_max else vals_max
 
 
-class DataScaler(torch.nn.Module):
+class GlobalScaler(torch.nn.Module):
     """Scale data to be in the range in a particular range.  Handles scaling on both input and output sides
     so if you can set the range of input and outputs separately."""
 
     def __init__(self, in_range=(0., 1.), out_range=(0., 1.), device=None, dtype=None):
         super().__init__()
-        # Default to range [-1, 1]
         self._in_range = Range(in_range, device=device, dtype=dtype)  # Input range
         self._out_range = Range(out_range, device=device, dtype=dtype)  # Output range
 
@@ -323,6 +326,25 @@ class DataScaler(torch.nn.Module):
         return self.scale(x)
 
 
+class ComponentWiseInputScaler(torch.nn.Module):
+
+    def __init__(self, inputs: torch.Tensor):
+        super().__init__()
+        xmin = torch.min(inputs, dim=0)[0]
+        xmax = torch.max(inputs, dim=0)[0]
+        delta = xmax - xmin
+
+        self._scale = 2 / delta
+        self._offset = (2 * xmin + delta) / delta
+
+        zeros = delta.isclose(torch.tensor(0., dtype=delta.dtype))
+        self._scale[zeros] = 1.
+        self._offset[zeros] = 0.
+
+    def forward(self, inp: torch.Tensor) -> torch.Tensor:
+        return inp * self._scale - self._offset
+
+
 class NeuralNetwork:
     """Neural network empirical potential model"""
 
@@ -364,30 +386,37 @@ class NeuralNetwork:
             self._device = device
         self._dtype = dtype
         self._bias = bias
-        self._vectorise_jacobian = vectorise_jacobian
+        self.vectorise_jacobian: bool = vectorise_jacobian
 
-        self._fingerprint_scaler = DataScaler(out_range=(-1., 1.), device=self._device, dtype=self._dtype)
-        self._energy_scaler = DataScaler(in_range=(-1., 1.), device=self._device, dtype=self._dtype)
+        self._fingerprint_scaler = GlobalScaler(out_range=(-1., 1.), device=self._device, dtype=self._dtype)
+        self._energy_scaler = GlobalScaler(in_range=(-1., 1.), device=self._device, dtype=self._dtype)
         self.loss_function = loss_function or LossFunction()
 
         self._network = None
 
+    def init(self, training_data: FittingData):
+        """Reinitialise the model.  This will throw away anything learned so far"""
+        self._create_network(training_data)
+
+    @property
+    def device(self):
+        return self._device
+
     def fit(
         self,
-        training_set: Union[Sequence[ase.Atoms], utils.FingerprintSet],
+        training_set: Union[Sequence[ase.Atoms], dat.FingerprintSet],
         max_epochs=200,
         progress_callback: Callable = None,
         learning_rate=5e-4,
         batchsize=16,
-    ):
+    ) -> Tuple[FittingData, LossFunction]:
+        if isinstance(training_set, FittingData):
+            training_data = training_set
+        else:
+            training_data = self.create_fitting_data(training_set, requires_grad=True)
+
         if self._network is None:
-            self._create_network(training_set.fingerprint_len)
-
-        training_data = self.create_fitting_data(training_set, requires_grad=True)
-
-        # Scale the inputs and outputs to match the ranges
-        norm_energies = training_data.get_normalised_energies()
-        self._energy_scaler.output.expand(norm_energies)
+            self._create_network(training_data)
 
         def stopping_function(epoch, _training, _loss):
             if epoch >= max_epochs:
@@ -395,7 +424,9 @@ class NeuralNetwork:
 
             return False  # Don't stop
 
-        optimiser = torch.optim.Adam(self._network.parameters(), lr=learning_rate)
+        optimiser = torch.optim.AdamW(self._network.parameters(), lr=learning_rate)
+        # optimiser = torch.optim.SGD(self._network.parameters(), lr=learning_rate, momentum=0.9)
+        # optimiser = torch.optim.LBFGS(self._network.parameters(), lr=learning_rate)
         loss = self._train(
             training_data,
             stopping_function,
@@ -405,7 +436,7 @@ class NeuralNetwork:
         )
         return training_data, loss
 
-    def create_fitting_data(self, fingerprint_set: utils.FingerprintSet, requires_grad=False):
+    def create_fitting_data(self, fingerprint_set: dat.FingerprintSet, requires_grad=False):
         return FittingData.from_fingerprint_set(
             fingerprint_set,
             device=self._device,
@@ -429,12 +460,14 @@ class NeuralNetwork:
         get_forces = self.loss_function.force_coeff != 0.
         while True:
             for batch in batches:
-                predictions = self.make_prediction(batch, get_forces=get_forces, create_graph=True)
-
                 optimiser.zero_grad()
+
+                predictions = self.make_prediction(batch, get_forces=get_forces, create_graph=True)
                 loss_result = self.loss_function.get_loss(predictions, batch)
-                loss = loss_result.total
-                loss.backward(retain_graph=True)
+
+                # Ask the optimiser to make a step
+                loss_result.total.backward(retain_graph=True)
+
                 optimiser.step()
 
             # Use no_grad to reduce memory footprint
@@ -460,59 +493,64 @@ class NeuralNetwork:
         return self.loss_function.get_loss(predictions, fitting_data)
 
     def make_prediction(self, fitting_data: FittingData, get_forces=False, create_graph=False) -> Predictions:
+        # pylint: disable=too-many-locals,too-many-nested-blocks
         # Join all fingerprints as this is much faster going through the network
         local_energies = self._network(fitting_data.fingerprints)
 
         forces = None
         if get_forces:
-            forces = []
             # Need to calculate the forces using the chain rule with the fingerprints derivative and the
             # neural network derivative
-            if self._vectorise_jacobian:
-                jac = functional.jacobian(
-                    self._network, fitting_data.fingerprints, create_graph=create_graph, vectorize=True
-                )
-                jacs = [jac[i, :, i, :] for i in range(len(fitting_data.fingerprints))]
+            if self.vectorise_jacobian:
+                if functorch is not None:
+                    # VMAP version, This came from a discussion with Richard Zou about.  See:
+                    # https://github.com/pytorch/pytorch/issues/42368
 
-                for fingerprint, derivatives, network_deriv in zip(
-                    fitting_data.fingerprints, fitting_data.derivatives, jacs
-                ):
-                    # Get the derivative of the energy wrt to input vector
-                    predicted_force = -torch.matmul(network_deriv, derivatives)
-                    forces.append(predicted_force)
-            else:
-                # Fallback, lower memory version
-                for fingerprint, derivatives in zip(fitting_data.fingerprints, fitting_data.derivatives):
-                    # Get the derivative of the energy wrt to input vector
-                    network_deriv = functional.jacobian(
-                        self._network, fingerprint, create_graph=create_graph, vectorize=True
+                    jacs = functorch.vmap(functorch.jacrev(self._network))(fitting_data.fingerprints)
+                    forces = -torch.matmul(jacs, fitting_data.derivatives)[:, 0, :]
+                else:
+                    forces = []
+
+                    jac = functional.jacobian(
+                        self._network, fitting_data.fingerprints, create_graph=create_graph, vectorize=True
                     )
-                    predicted_force = -torch.matmul(network_deriv, derivatives)
-                    forces.append(predicted_force)
+                    jacs = [jac[i, :, i, :] for i in range(len(fitting_data.fingerprints))]
 
-            # VMAP version, doesn't work yet but may be fixed soon and would almost certainly be fast and relatively
-            # low-memory.  See:
-            # https://github.com/pytorch/pytorch/issues/42368
-            # def get_jacobian(fingerprint):
-            #     return functional.jacobian(self._network, fingerprint, create_graph=False)
-            #
-            # jac = torch.vmap(get_jacobian)(tuple(fitting_data.fingerprints))
-            # jacs = [jac[i, :, i, :] for i in range(len(fitting_data.fingerprints))]
-            #
-            # for fingerprint, derivatives, network_deriv
-            #   in zip(fitting_data.fingerprints, fitting_data.derivatives, jacs):
-            #     # Get the derivative of the energy wrt to input vector
-            #     predicted_force = -torch.matmul(network_deriv, derivatives)
-            #     forces.append(predicted_force)
+                    for _fingerprint, derivatives, network_deriv in zip(
+                        fitting_data.fingerprints, fitting_data.derivatives, jacs
+                    ):
+                        # Get the derivative of the energy wrt to input vector
+                        predicted_force = -torch.matmul(network_deriv, derivatives)
+                        forces.append(predicted_force)
 
-            forces = torch.cat(forces)
+                    forces = torch.cat(forces)
+            else:
+                # Fallback: lower memory, but slow, version
+                forces = []
+
+                for system in fitting_data.iter_systemwise():
+                    natoms = system.num_atoms
+                    denergy_dphi = tuple(
+                        functional.jacobian(self._network, fingerprint, create_graph=create_graph, strict=True)[0]
+                        for fingerprint in system.fingerprints
+                    )
+                    # Calculate the force on each atom
+                    for i in range(natoms):
+                        total_force = torch.zeros(3, device=self.device)
+                        for j in range(natoms):
+                            if not torch.all(system.derivatives[j, i, :, :] == 0):
+                                total_force += -torch.matmul(system.derivatives[j, i, :, :], denergy_dphi[j])
+                        forces.append(total_force)
+
+                forces = torch.stack(forces)
 
         return Predictions(fitting_data.num_atoms, fitting_data.index, local_energies=local_energies, forces=forces)
 
-    def _create_network(self, input_size):
+    def _create_network(self, training_data: FittingData):
         # Create the network
-        sequence = []
-        prev_size = input_size
+        sequence = [ComponentWiseInputScaler(training_data.fingerprints)]
+
+        prev_size = training_data.fingerprint_len
         # Create the fully connected (hidden) layers
         for size in self._hidden:
             sequence.append(torch.nn.Linear(prev_size, size, bias=self._bias))
@@ -521,6 +559,9 @@ class NeuralNetwork:
         # Add the output layer
         sequence.extend((torch.nn.Linear(prev_size, 1, bias=self._bias), self._energy_scaler))
 
+        # Scale the energy outputs to match the ranges
+        self._energy_scaler.output.expand(training_data.get_normalised_energies())
+
         self._network = torch.nn.Sequential(*sequence)
         self._network.to(self._device, self._dtype)
 
