@@ -1,9 +1,7 @@
 # -*- coding: utf-8 -*-
-import argparse
 import functools
 import collections
 import logging
-import random
 from typing import List
 
 import numpy as np
@@ -16,10 +14,9 @@ from . import atomic
 from . import base_moments
 from . import exceptions
 from . import fingerprinting
+from . import mathutil
 from . import optimisers
 
-__all__ = ('Decoder',)
-
 _LOGGER = logging.getLogger(__name__)
 
 StructureOptimisationResult = collections.namedtuple('StructureOptimisationResult', 'success atoms message rmsd')
@@ -33,7 +30,7 @@ def find_clusters(spec, num_clusters: int, **kwargs) -> np.ndarray:
 
 @find_clusters.register(base_moments.Moments)
 def _(moments: base_moments.Moments, num_clusters: int, query: base_moments.ReconstructionQuery,
-      fingerprinter: fingerprinting.MomentInvariantsDescriptor) -> np.ndarray:
+      descriptor: fingerprinting.MomentInvariantsDescriptor) -> np.ndarray:
     """Find clusters from moments.  This will take the moments and reconstruct values on a grid
     which will be used for the actual cluster determination"""
     # Calculate the grid values
@@ -60,152 +57,6 @@ def _(grid, num_clusters: int) -> np.ndarray:
     return kmeans.cluster_centers_
 
 
-def find_maximum(moments: base_moments.Moments, initial: np.ndarray) -> np.ndarray:
-    """Find the maximum of a set of moments"""
-
-    def calc(pos: np.ndarray):
-        return -moments.value_at(pos)
-
-    res = optimize.minimize(calc, initial)
-    return res.x
-
-
-def find_peaks(
-    num_peaks: int,
-    moments: base_moments.Moments,
-    descriptor: fingerprinting.MomentInvariantsDescriptor,
-    query: base_moments.ReconstructionQuery = None,
-    grid_size=31,
-):
-    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)
-    current_grid = moments.reconstruct(query, zero_outside_domain=True)
-    return find_peaks_from_grid(num_peaks, current_grid, descriptor, query)
-
-
-class FoundPeaks:
-
-    def __init__(self, initial: base_moments.Moments):
-        self._initial = initial
-        self._moments = []
-        self._atom_positions = []
-
-    @property
-    def positions(self) -> List[np.ndarray]:
-        return self._atom_positions
-
-    @property
-    def moments(self) -> List[base_moments.Moments]:
-        return self._moments
-
-    def append(self, pos: np.ndarray, moments: base_moments.Moments):
-        self._atom_positions.append(pos)
-        self._moments.append(moments)
-
-    def remaining_value(self, pos: np.ndarray) -> float:
-        initial = self._initial.value_at(pos)
-        for atom_moments in self._moments:
-            initial -= atom_moments.value_at(pos)
-        return initial
-
-    def find_next_maximum(self, start_pt: np.ndarray):
-        res = optimize.minimize(self.remaining_value, start_pt)
-        return res.x
-
-
-def find_peaks_from_maxima(
-    num_peaks: int,
-    moments: base_moments.Moments,
-    descriptor: fingerprinting.MomentInvariantsDescriptor,
-    query: base_moments.ReconstructionQuery = None,
-    grid_size=31,
-):
-    # pylint: disable=too-many-locals
-    scale = 1.0
-    if descriptor.scaler is not None:
-        scale = 1. / descriptor.cutoff
-
-    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)
-    peaks = FoundPeaks(moments)
-    current_grid = moments.reconstruct(query, zero_outside_domain=True)
-    outside = get_buffer_indices(query, current_grid)
-    current_grid[outside] = 0.
-
-    for _ in range(num_peaks):
-        # Find the index of the maximum value in the current grid
-        max_idx = current_grid.argmax()
-
-        # Get that position in the grid
-        guess = query.points[max_idx]
-        # Now find the local maximum
-        moments_pos = peaks.find_next_maximum(guess)
-
-        # Build an atoms collection with a single atom at that position
-        single_atom = atomic.AtomsCollection(1, positions=[moments_pos / scale], numbers=[1.])
-
-        # Get the moments so we can subtract this from the grid
-        single_moments = descriptor.get_moments(single_atom, preprocess=False)
-        peaks.append(moments_pos, single_moments)
-
-        # Subtract off the grid
-        # Get the grid for just that atom on its own
-        atom_grid = single_moments.reconstruct(query, zero_outside_domain=True)
-        # Subtract off the single atom grid
-        current_grid -= atom_grid
-
-        current_grid[outside] = 0.
-
-    return peaks.positions / scale
-
-
-def find_peaks_from_grid(
-    num_peaks: int,
-    grid,
-    descriptor: fingerprinting.MomentInvariantsDescriptor,
-    query: base_moments.ReconstructionQuery,
-    exclude_radius=0.51,
-):
-    # pylint: disable=too-many-locals
-    subtract_signal = True
-    scale = 1.0 if descriptor.cutoff is None else 1. / descriptor.cutoff
-
-    current_grid = grid.copy()
-
-    exclude = exclude_radius * scale
-    outside = get_buffer_indices(query, current_grid)
-
-    mask = np.zeros(current_grid.shape, dtype=bool)
-    mask[outside] = True
-
-    found_positions = []
-    for _ in range(num_peaks):
-        # Find the index of the maximum value in the current grid
-        max_idx = current_grid.argmax()
-
-        # Get that position in the grid
-        atom_pos = query.points[max_idx]
-        found_positions.append(atom_pos / scale)
-
-        if subtract_signal:
-            # Build an atoms collection with a single atom at that position
-            single_atom = atomic.AtomsCollection(1, positions=[atom_pos / scale], numbers=[1.])
-
-            # Get the moments so we can subtract this from the grid
-            single_moments = descriptor.get_moments(single_atom, preprocess=False)
-
-            # Get the grid for just that atom on its own
-            atom_grid = single_moments.reconstruct(query, zero_outside_domain=True)
-
-            # Subtract off the single atom grid
-            current_grid -= atom_grid
-
-        # Now mask off the gridpoints associated with this atom
-        mask |= get_surrounding_gridpoints(query, atom_pos, exclude)
-
-        current_grid[mask] = 0.
-
-    return np.array(found_positions)
-
-
 def get_surrounding_gridpoints(
     query: base_moments.ReconstructionQuery, atom_pos: np.array, radius: float
 ) -> np.ndarray:
@@ -328,89 +179,6 @@ DecoderResult = collections.namedtuple(
 )
 
 
-class Decoder:
-    """This class decodes a structure from a set of moment invariants"""
-
-    def __init__(
-        self,
-        descriptor: fingerprinting.MomentInvariantsDescriptor,
-        query: base_moments.ReconstructionQuery,
-    ):
-        self._descriptor = descriptor
-        self._query = query
-
-        self._optimiser = optimisers.StructureOptimiser()
-        self._moments_optimiser = optimisers.MomentsOptimiser()
-        self._structure_optimiser = optimisers.StructureOptimiser()
-
-    def decode(self, fingerprint: np.ndarray, num_atoms: int, verbose=False, get_trajectory=False) -> DecoderResult:
-        decode_result = argparse.Namespace()
-
-        # Generate an initial configuration using a structure with randomly placed atoms
-        initial_atoms = self.create_random_atoms(num_atoms)
-
-        # Reconstruct the moments
-        result = self._moments_optimiser.optimise(
-            self._descriptor.invariants, target=fingerprint, initial=self._descriptor.get_moments(initial_atoms)
-        )
-
-        decode_result.moments_reconstruction = result
-
-        if verbose:
-            print(f'm: {result.rmsd}, ', end='')
-
-        initial_atoms = self.create_initial_atoms(result.value, num_atoms)
-
-        # Now let's get optimise the atomic configuration to be consistent with the reconstructed moments
-        result = self._structure_optimiser.optimise(
-            self._descriptor,
-            initial=initial_atoms,
-            target=result.value,  # Target the found moments
-            get_trajectory=get_trajectory,
-        )
-
-        decode_result.initial_reconstruction = result
-
-        if verbose:
-            print(f'sm: {result.rmsd}, ', end='')
-
-        # Finally, optimise the found structure wrt to the fingerprint
-        result = self._structure_optimiser.optimise(
-            self._descriptor,
-            initial=result.value,
-            target=fingerprint,  # Target the fingerprint
-            get_trajectory=get_trajectory,
-        )
-
-        if verbose:
-            print(f'sf: {result.rmsd}, ', end='')
-
-        decode_result.atoms_reconstruction = result
-        decode_result.success = result.success
-        decode_result.message = result.message
-        decode_result.value = result.value
-        decode_result.rmsd = result.rmsd
-
-        return DecoderResult(**decode_result.__dict__)
-
-    def create_random_atoms(self, num: int) -> atomic.AtomsCollection:
-        """Create a random atoms configuration with the number of atoms passed"""
-        return atomic.random_atom_collection_in_sphere(
-            num,
-            radius=self._descriptor.cutoff,
-            numbers=random.choices(self._descriptor.species, k=num),
-            centre=True,
-        )
-
-    def create_initial_atoms(self, moments: base_moments.Moments, num: int) -> atomic.AtomsCollection:
-        # atoms = find_atoms(num, moments, self._descriptor, self._query)
-        # atoms.numbers[:] = random.choices(self._descriptor.species, k=num)
-        # return atoms
-
-        peaks = find_peaks(num, moments, self._descriptor, self._query)
-        return atomic.AtomsCollection(num, peaks, random.choices(self._descriptor.species, k=num))
-
-
 def merge_atoms(system: atomic.AtomsCollection, dist_threshold=0.2):
     # Proceed to merging of atoms
     dists = distance.cdist(system.positions, system.positions)
@@ -518,20 +286,54 @@ def get_best_rms(
 def find_iteratively(
     descriptor: fingerprinting.MomentInvariantsDescriptor,
     fingerprint: np.ndarray,
-    num_atoms: int,
     initial: atomic.AtomsCollection,
     find_species=False,
-    verbose=False,
     min_rmsd=1e-7,
     max_iters=6,
-    grid_query=None
-):
-    # pylint: disable=too-many-locals
-    # Initialisation
-    moments_optimiser = optimisers.MomentsOptimiser()
-    structure_optimiser = optimisers.StructureOptimiser()
-    atoms = initial
+    grid_query=None,
+    structure_optimiser=None,
+    minsep=0.55,
+    verbose=False,
+) -> optimisers.StructureOptimisationResult:
+    """
+    Find a set of atoms from a fingerprint using an iterative approach.
 
+    This algorithm proceeds in iterations as follows:
+        1. Gradient descent to find the moments from the invariants (using `initial` atoms as the starting point)
+        2. From the found moments:
+            i) Identify the peaks in the grid sampled from the reconstructed moments
+            ii) Create a collection of atoms and gradient descent to find the minimum of the atomic degrees of freedom
+                wrt the moments
+        3. Take the found atoms collection and perform gradient descent of the atomic degrees of freedom wrt to the
+           original fingerprint.
+
+    This is then repeated until either the RMSD between the fingerprint of the found atoms collection and the target
+    drops below the passed threshold or the max number of iterations is reached.
+
+    :param descriptor: the descriptor to use
+    :param fingerprint: the target fingerprint
+    :param initial: the initial atoms collection to start from
+    :param find_species: if True, will look gradient descent atomic species as well as atomic coordinates
+    :param min_rmsd: the fingerprint stopping threshold
+    :param max_iters: the maximum number of iterations
+    :param grid_query: an optional grid query, this is used when supersampling the found moments and if this function
+        is being called many times using the same descriptor it is much faster to create the query externally and pass
+        it in each time
+    :param structure_optimiser: an optional structure optimiser to use
+    :param minsep: the minimum separation between atoms
+    :param verbose: if True, will print information about progress
+    :return: the result from the last structure optimisation
+    """
+    # pylint: disable=too-many-locals, too-many-branches
+    # Initialisation
+    num_atoms = initial.num_atoms
+    moments_optimiser = optimisers.MomentsOptimiser()
+    if structure_optimiser is None:
+        structure_optimiser = optimisers.StructureOptimiser()
+        if minsep:
+            structure_optimiser.separation_force = atomic.SeparationForce(epsilon=1e-1, cutoff=minsep, power=2)
+
+    atoms = initial
     mask = None
     if not find_species:
         # Fix the species numbers
@@ -546,51 +348,46 @@ def find_iteratively(
             # Create the reconstruction query the first time
             grid_query = moments.create_reconstruction_query(moments.get_grid(31), moments.max_order)
 
+        if verbose:
+            print(f'{i}: Finding moments from fingerprint...', end='')
         result = moments_optimiser.optimise(
             invariants_fn=descriptor.invariants,
             target=fingerprint,
             initial=moments,
-            verbose=False,
-            # cost_tol=1e-5,
+            verbose=(verbose == 'high'),
         )
         moments = result.value
+        if verbose:
+            print(f'rmsd {result.rmsd}')
 
         if verbose:
-            print(f'{i} moms->fingerprint: {result.rmsd}')
-
-        # Find the peaks and create the corresponding collection of atoms
-        peaks = find_peaks(num_atoms, moments, descriptor, grid_query)
-        atoms = atomic.AtomsCollection(num_atoms, peaks, numbers=initial.numbers)
-
-        result = structure_optimiser.optimise(
+            print(f'{i}: Finding atoms from moments...', end='')
+        result = find_atoms_from_moments(
             descriptor,
-            target=moments,
-            initial=atoms,
+            moments,
+            num_atoms,
+            numbers=atoms.numbers,
+            minsep=minsep,
             mask=mask,
-            verbose=False,
+            grid_query=grid_query,
+            structure_optimiser=structure_optimiser,
+            verbose=(verbose == 'high')
         )
+        atoms = result.value
+        if verbose:
+            print(f'rmsd {result.rmsd}')
 
         if verbose:
-            print(f'{i} atoms->moms: {result.rmsd}')
-
+            print(f'{i}: Finding atoms from fingerprint...', end='')
         result = structure_optimiser.optimise(
             descriptor,
             target=fingerprint,
-            initial=result.value,
+            initial=atoms,
             mask=mask,
-            verbose=False,
+            verbose=(verbose == 'high'),
         )
-
-        # result = structure_optimiser.optimise(
-        #     descriptor,
-        #     target=fingerprint,
-        #     initial=atoms,
-        #     mask=mask,
-        #     verbose=False,
-        # )
-
         if verbose:
-            print(f'{i} atoms->fingerprint: {result.rmsd}')
+            print(f'rmsd {result.rmsd}')
 
         if find_species:
             # Take the current result, fix the species and allow positions to vary
@@ -599,16 +396,17 @@ def find_iteratively(
             pos_mask = atoms.get_mask()
             pos_mask.numbers = atoms.numbers
 
+            if verbose:
+                print(f'{i}: Optimising atomic positions wrt fingerprint...', end='')
             result = structure_optimiser.optimise(
                 descriptor,
                 target=fingerprint,
                 initial=atoms,
                 mask=pos_mask,
-                verbose=False,
+                verbose=(verbose == 'high'),
             )
-
             if verbose:
-                print(f'{i} atom pos->fingerprint: {result.rmsd}')
+                print(f'rmsd {result.rmsd}')
 
         if result.rmsd < min_rmsd:
             break
@@ -616,3 +414,136 @@ def find_iteratively(
         atoms = result.value
 
     return result
+
+
+# Interface for atoms finder:
+# (descriptor, moments, num_atoms: int, *, verbose: Bool)
+
+
+def find_atoms_from_moments(
+    descriptor: fingerprinting.MomentInvariantsDescriptor,
+    moments,
+    num_atoms: int,
+    *,
+    numbers=1.,
+    mask=None,
+    grid_query=None,
+    structure_optimiser=None,
+    minsep=0.51,
+    verbose=False,
+):
+    # Find the peaks and create the corresponding collection of atoms
+    if verbose:
+        print('Finding peaks...', end='')
+    peaks = find_peaks(descriptor, moments, num_atoms, query=grid_query, subtract_signal=True, exclude_radius=minsep)
+    atoms = atomic.AtomsCollection(num_atoms, peaks, numbers=numbers)
+    if verbose:
+        print(f'found {atoms.num_atoms}')
+
+    if structure_optimiser is None:
+        structure_optimiser = optimisers.StructureOptimiser()
+        structure_optimiser.separation_force = atomic.SeparationForce(epsilon=1e-2, cutoff=minsep, power=1)
+    result = structure_optimiser.optimise(
+        descriptor,
+        target=moments,
+        initial=atoms,
+        mask=mask,
+        verbose=verbose,
+    )
+
+    return result
+
+
+def find_maximum(
+    descriptor: fingerprinting.MomentInvariantsDescriptor,
+    moments: base_moments.Moments,
+    num_samples=128,
+) -> optimize.OptimizeResult:
+    """Given a set of moments this function will use a function minimisation algorithm to find
+    the global maximum of the moments reconstruction
+
+    :param descriptor: the descriptor that created the moments
+    :param moments: the moments
+    :param num_samples: the number of samples to use for the first pass of the global maximum search
+    """
+    # pylint: disable=invalid-name
+    cutoff = descriptor.cutoff
+
+    def adapter(moms, pt):
+        return -moms.value_at(np.array([mathutil.sph2cart(pt)]))
+
+    res = optimize.shgo(
+        lambda pt: adapter(moments, pt),
+        bounds=[(0.0, 1.), (0., np.pi), (0., 2 * np.pi)],
+        n=num_samples,
+        sampling_method='sobol'
+    )
+    # Convert coordinates back to cartesian and scale
+    res.update(dict(x=mathutil.sph2cart(res.x) / cutoff, xl=mathutil.sph2cart(res.x) / cutoff))
+    return res
+
+
+def find_peaks(
+    descriptor: fingerprinting.MomentInvariantsDescriptor,
+    moments: base_moments.Moments,
+    num: int,
+    *,
+    query: base_moments.ReconstructionQuery = None,
+    grid_size=31,
+    exclude_radius=0.51,
+    subtract_signal=True
+):
+    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)
+    current_grid = moments.reconstruct(query, zero_outside_domain=True)
+    return find_peaks_from_grid(
+        descriptor, num, current_grid, query, exclude_radius=exclude_radius, subtract_signal=subtract_signal
+    )
+
+
+def find_peaks_from_grid(
+    descriptor: fingerprinting.MomentInvariantsDescriptor,
+    num_peaks: int,
+    grid_values: np.ndarray,
+    query: base_moments.ReconstructionQuery,
+    exclude_radius=0.51,
+    subtract_signal=True
+):
+    # pylint: disable=too-many-locals
+    scale = 1.0 if descriptor.cutoff is None else 1. / descriptor.cutoff
+
+    current_grid = grid_values.copy()
+
+    exclude = exclude_radius * scale
+    outside = get_buffer_indices(query, current_grid)
+
+    mask = np.zeros(current_grid.shape, dtype=bool)
+    mask[outside] = True
+
+    found_positions = []
+    for _ in range(num_peaks):
+        # Find the index of the maximum value in the current grid
+        max_idx = current_grid.argmax()
+
+        # Get that position in the grid
+        atom_pos = query.points[max_idx]
+        found_positions.append(atom_pos / scale)
+
+        if subtract_signal:
+            # Build an atoms collection with a single atom at that position
+            single_atom = atomic.AtomsCollection(1, positions=[atom_pos / scale], numbers=[1.])
+
+            # Get the moments so we can subtract this from the grid
+            single_moments = descriptor.get_moments(single_atom, preprocess=False)
+
+            # Get the grid for just that atom on its own
+            atom_grid = single_moments.reconstruct(query, zero_outside_domain=True)
+
+            # Subtract off the single atom grid
+            current_grid -= atom_grid
+
+        # Now mask off the gridpoints associated with this atom
+        mask |= get_surrounding_gridpoints(query, atom_pos, exclude)
+
+        current_grid[mask] = 0.
+
+    return np.array(found_positions)
