@@ -2,42 +2,81 @@
 """Module for calculating Zernike moments"""
 
 import collections
+import copy
 import logging
 import itertools
 import math
 import numbers
 import functools
 import random
-from typing import Union, Tuple, Iterator, Dict, Optional
+from typing import Union, Tuple, Iterator
 
 import numpy as np
 from numpy import ma
-import scipy
-import scipy.special
+from scipy import special
 
 from . import base_moments
 from . import functions
+from . import mathutil
 from . import geometric
-from .utils import even, inclusive
+from . import polynomials
+from . import sph
+from .utils import even, odd, inclusive
 
 _LOGGER = logging.getLogger(__name__)
 
 # pylint: disable=invalid-name
 
-__all__ = 'ZernikeMoments', 'ZernikeMomentCalculator', 'from_deltas', 'from_gaussians'
+__all__ = 'ZernikeMoments', 'ZernikeMomentsCalculator', 'ZernikeMomentCalculator', 'from_deltas', 'from_gaussians'
+
+NORMALISTAION = 3. / (4. * np.pi)
 
 
-def from_deltas(max_order: int, positions: np.array, weights: Union[numbers.Number, np.array] = 1.) -> 'ZernikeMoments':
+def from_deltas(
+    n_max: int,
+    positions: np.array,
+    weights: Union[numbers.Number, np.array] = 1.,
+    l_max: int = None
+) -> 'ZernikeMoments':
     """Create a set of Zernike moments from a collection of delta functions with optional weights
 
-    :param max_order: the order of Zernike moments to calculate to
+    :param n_max: the maximum order of Zernike functions to use
     :param positions: the Cartesian positions of the delta functions
     :param weights: the weights of the delta functions, can be a scalar or np.array of the same
         length as positions
+    :param l_max: the maximum spherical harmonic angular frequency, l, to go up to
     """
     _domain_check(positions)
-    geom_moments = geometric.from_deltas(max_order, positions, weights)
-    return from_geometric_moments(max_order, geom_moments)
+    geom_moments = geometric.from_deltas(n_max, positions, weights)
+    return from_geometric_moments(n_max, geom_moments, l_max)
+
+
+def from_deltas_direct(
+    nmax: int, positions: np.ndarray, weights: Union[numbers.Number, np.array] = 1., lmax=None
+) -> 'ZernikeMoments':
+    """This is the method used to calcualte fingerprints by AMP (at least when not using gradients).
+
+    There is some discreptancy between this and the method from geometric moments proposed by Novotni and Klein, however
+    it amounts to a simple rotation as the invariants are still the same.  We have this here, mainly so that
+    we can easily compare with AMP code.
+    """
+    norm = NORMALISTAION
+    zernike_moms = ZernikeMoments(n_max=nmax, l_max=lmax)
+
+    if len(positions) > 0:
+        spherical = mathutil.cart2sph(positions.T)  # Spherical coordinates
+        for n in inclusive(nmax):
+            for l in inclusive(n):
+                if odd(n - l):
+                    continue
+
+                for m in inclusive(-l, l):
+                    vals = norm * zernike_poly(n, l, m, spherical, condon_shortley=False).conjugate()
+                    zernike_moms.array[n, l, m] = np.sum(weights * vals)
+    else:
+        zernike_moms.array.fill(0.)
+
+    return zernike_moms
 
 
 def from_gaussians(
@@ -62,14 +101,15 @@ def from_gaussians(
 
 
 def from_geometric_moments(
-    max_order: int, geom_moments: Union[geometric.GeometricMoments, np.array]
+    n_max: int, geom_moments: Union[geometric.GeometricMoments, np.array], l_max: int = None
 ) -> 'ZernikeMoments':
     """Create a set of Zernike moments from a set of geometric moments
 
-    :param max_order: the order of Zernike moments to calculate to
+    :param n_max: the maximum order of Zernike functions to use
     :param geom_moments: the geometric moments
+    :param l_max: the maximum spherical harmonic angular frequency, l, to go up to
     """
-    return ZernikeMomentCalculator(max_order)(geom_moments)
+    return ZernikeMomentsCalculator(n_max, l_max)(geom_moments)
 
 
 class ZernikeReconstructionQuery(base_moments.ReconstructionQuery):
@@ -96,7 +136,7 @@ class ZernikeReconstructionQuery(base_moments.ReconstructionQuery):
     def valid_idxs(self):
         """Get the indexes of points that are within the domain"""
         if self._valid_idxs is None:
-            self._valid_idxs = ZernikeMoments._get_indices_in_domain(self.points)
+            self._valid_idxs = _get_indices_in_domain(self.points)
         return self._valid_idxs
 
 
@@ -107,12 +147,14 @@ ZernikeIndex = collections.namedtuple('ZernikeIndex', 'n l m')
 class ZernikeMoments(base_moments.Moments):
     """A container class for calculated Zernike moments"""
 
+    # pylint: disable=too-many-public-methods
+
     @classmethod
     def from_indexed(cls, indexed, max_order: int, dtype=None) -> 'ZernikeMoments':
         dtype = dtype or type(indexed[0, 0, 0])
         moments = cls(max_order, dtype=dtype)
         for (n, l, m) in moments.iter_indices():
-            moments._set_moment(n, l, m, indexed[n, l, m], set_minus_m=False)
+            cls._set_moment(moments, n, l, m, indexed[n, l, m], set_minus_m=False)
         return moments
 
     @classmethod
@@ -138,25 +180,21 @@ class ZernikeMoments(base_moments.Moments):
 
         return moms
 
-    def __init__(self, n_max: int, dtype=complex):
+    def __init__(self, n_max: int, l_max=None, dtype=complex):
         """Construct a Zernike moments object
         """
-        self._max_n = n_max
+        self._indices = index_traits(n_max, l_max)
         self._dtype = dtype
 
-        moments = np.empty((n_max + 1, n_max + 1, 2 * n_max + 1), dtype=dtype)
+        moments = sph.create_array(self._indices, dtype=dtype)
+        self._moments = moments
         # Fill it with NaN so that it's easy to spot any values that haven't been set but should be
-        moments.fill(np.nan)
+        self._moments.fill(np.nan)
 
-        # Mask everything and then unmask the ones we want
-        mask = np.ones(moments.shape, dtype=np.int)
-        idx_array = np.empty(moments.shape, dtype=np.int)
-        for i, indices in enumerate(iter_indices(n_max, redundant=True)):
-            mask[indices] = 0
-            idx_array[indices] = i
-
-        self._moments = ma.array(moments, mask=mask)
-        self._idx_array = ma.array(idx_array, mask=mask)
+        # Create an index array that allows us to map from a multi-index like [5, 3, -1] to a linear index
+        idx_array = ma.masked_array(np.empty(moments.shape, dtype=int), mask=moments.mask)
+        idx_array[~moments.mask] = np.arange(moments.count())
+        self._idx_array = idx_array
 
     def __eq__(self, other: 'ZernikeMoments'):
         """Test if another set of moments are equal to this one"""
@@ -165,6 +203,23 @@ class ZernikeMoments(base_moments.Moments):
 
         return np.all(self.vector == other.vector)
 
+    def __add__(self, other: 'ZernikeMoments') -> 'ZernikeMoments':
+        if not isinstance(other, ZernikeMoments):
+            raise ValueError(f'Expected ZernikeMoments, got {type(other)}')
+
+        moms = copy.deepcopy(self)
+        moms.array.data[:] += other.array.data[:]
+
+        return moms
+
+    def __sub__(self, other: 'ZernikeMoments') -> 'ZernikeMoments':
+        if not isinstance(other, ZernikeMoments):
+            raise ValueError(f'Expected ZernikeMoments, got {type(other)}')
+
+        moms = copy.deepcopy(self)
+        moms.array.data[:] -= other.array.data[:]
+        return moms
+
     @property
     def dtype(self):
         return self._dtype
@@ -178,18 +233,26 @@ class ZernikeMoments(base_moments.Moments):
 
     @property
     def max_order(self) -> int:
-        return self._max_n
+        return self._indices.n.max
+
+    @property
+    def n_max(self) -> int:
+        return self._indices.n.max
+
+    @property
+    def l_max(self) -> int:
+        return self._indices.l.max
 
     def get_builder(self, mask: 'Optional[ZernikeMoments]' = None):
-        return ZernikeMomentsBuilder(self._max_n, mask)
+        return ZernikeMomentsBuilder(self.max_order, mask=mask)
 
     def get_mask(self) -> 'ZernikeMoments':
-        moms = ZernikeMoments(self._max_n, dtype=object)
+        moms = ZernikeMoments(self.n_max, self.l_max, dtype=object)
         moms.array.fill(None)
         return moms
 
     def randomise(self, num_range=(-5, 5), indices=(None, None, None)):
-        for n, l, m in iter_indices_extended(indices, self._max_n, redundant=False):
+        for n, l, m in iter_indices_extended(indices, self.n_max, redundant=False):
             mom = random.uniform(*num_range)
             if m != 0:
                 # Only l != 0 moments have a complex part
@@ -204,7 +267,7 @@ class ZernikeMoments(base_moments.Moments):
         return self._moments.compressed()
 
     def __getitem__(self, item):
-        return self.moment(*item)
+        return self._moments[item[0], item[1], item[2]]
 
     def __setitem__(self, key: Union[int, Tuple], value):
         self._set_moment(*key, value)
@@ -220,8 +283,8 @@ class ZernikeMoments(base_moments.Moments):
                 try:
                     if value.imag > 1e-9:
                         logging.warning(
-                            f'Trying to set value of moment {n},{l},{m} to {value}, '
-                            f'however these moment should always be real so discarding imaginary'
+                            'Trying to set value of moment %i,%i,%i to %d, '
+                            'however these moment should always be real so discarding imaginary', n, l, m, value
                         )
                     value = np.real(value)
                 except AttributeError:
@@ -238,8 +301,12 @@ class ZernikeMoments(base_moments.Moments):
         for indices in self.iter_indices(redundant=redundant):
             yield indices, self[indices]
 
+    @property
+    def indices(self):
+        return self._indices
+
     def iter_indices(self, redundant=True):
-        yield from iter_indices(max_order=self._max_n, redundant=redundant)
+        yield from iter_indices(n_max=self.n_max, l_max=self.l_max, redundant=redundant)
 
     def linear_index(self, index: base_moments.Index) -> int:
         """Given a triple of Zernike moment indices this will return the corresponding linearly ordered index"""
@@ -260,10 +327,13 @@ class ZernikeMoments(base_moments.Moments):
         :param x: the point to get the value at
         :param order: the maximum order to go up to (defaults to the order of these moments)
         """
-        order = order or self._max_n
+        order = order or self.n_max
         return self.reconstruct(self.create_reconstruction_query(x, order), order, zero_outside_domain=False)
 
-    def reconstruct(self, query: ZernikeReconstructionQuery, order=None, zero_outside_domain=True):
+    def reconstruct(self,
+                    query: ZernikeReconstructionQuery,
+                    order=None,
+                    zero_outside_domain=True) -> Union[float, np.ndarray]:
         """Given a reconstruction query this will return the values of the function at the specified grid values"""
         order = order if order is not None else query.max_order
         moments = query.moments
@@ -280,16 +350,15 @@ class ZernikeMoments(base_moments.Moments):
         else:
             calculated_values = values
 
-        for n, l, m in iter_indices(max_order=order, redundant=False):
-            omega = self.moment(n, l, m)
-            z = sum_chi_nlm(n, l, m, moments)
+        for n, l, m in iter_indices(n_max=order, redundant=False):
+            omega = self._moments[n, l, m]
+            z = ZernikeMomentsCalculator.change_of_basis(n, l, m)(moments)
             calculated_values += omega * z
 
             if m != 0:
                 # Now do the symmetric -m part
                 z_conj = (-1)**m * z.conjugate()
-
-                calculated_values += z_conj * self.moment(n, l, -m)
+                calculated_values += z_conj * self._moments[n, l, -m]
 
         if zero_outside_domain:
             np.put(values, query.valid_idxs, calculated_values)
@@ -297,8 +366,15 @@ class ZernikeMoments(base_moments.Moments):
         return values.real
 
     def visualise(
-        self, viewer='plotly_widget', query: ZernikeReconstructionQuery = None, max_order=None, num_points=31
+        self,
+        viewer='plotly_widget',
+        query: ZernikeReconstructionQuery = None,
+        max_order=None,
+        num_points=31,
+        square=False
     ):
+        # pylint: disable=too-many-locals
+
         if viewer.startswith('plotly_widget'):
             volume_plot = viewer.endswith('volume')
 
@@ -314,6 +390,8 @@ class ZernikeMoments(base_moments.Moments):
                 grid_points = query.points
 
             grid_values = self.reconstruct(query, max_order, zero_outside_domain=True)
+            if square:
+                grid_values = grid_values**2
 
             default_isomin = (grid_values.max() - grid_values.min()) * 0.6 + grid_values.min()
             default_isomax = grid_values.max()
@@ -340,12 +418,13 @@ class ZernikeMoments(base_moments.Moments):
 
             default_opacity = 0.05 if volume_plot else 0.4
 
+            # pylint: disable=unused-variable
             @ipywidgets.interact(
                 opacity=(0., 1., 0.01),
                 isomin=(grid_values.min(), grid_values.max(), 0.5),
                 isomax=(grid_values.min(), grid_values.max(), 0.5)
             )
-            def update(opacity=default_opacity, isomin=default_isomin, isomax=default_isomax):  # pylint: disable=unused-variable
+            def update(opacity=default_opacity, isomin=default_isomin, isomax=default_isomax):
                 with widget.batch_update():
                     isosurface = widget.data[0]
                     isosurface.opacity = opacity
@@ -380,11 +459,11 @@ class ZernikeMoments(base_moments.Moments):
 
         return moments
 
-    @staticmethod
-    def _get_indices_in_domain(points: np.ndarray) -> np.ndarray:
-        """Calculate the lengths squared and get the corresponding indexes"""
-        length_sq = (points**2).sum(axis=1)
-        return np.argwhere(length_sq < 1)
+
+def _get_indices_in_domain(points: np.ndarray) -> np.ndarray:
+    """Calculate the lengths squared and get the corresponding indexes"""
+    length_sq = (points**2).sum(axis=1)
+    return np.argwhere(length_sq < 1)
 
 
 class ZernikeMomentsBuilder(functions.Function):
@@ -392,9 +471,10 @@ class ZernikeMomentsBuilder(functions.Function):
     input_type = np.ndarray
     supports_jacobian = True
 
-    def __init__(self, n_max: int, mask: Optional[ZernikeMoments] = None):
+    def __init__(self, n_max: int, l_max: int = None, mask: ZernikeMoments = None):
         super().__init__()
         self._n_max = n_max
+        self._l_max = n_max if l_max is None else l_max
         self._mask = mask
         self._indices = None
 
@@ -410,7 +490,7 @@ class ZernikeMomentsBuilder(functions.Function):
     def inverse(self) -> 'ZernikeMomentsBuilder.Inverse':
         return ZernikeMomentsBuilder.Inverse(self)
 
-    def evaluate(self, state: np.ndarray, get_jacobian=False):
+    def evaluate(self, state: np.ndarray, *, get_jacobian=False):  # pylint: disable=arguments-differ
         indices_list = tuple(self.iter_indices())
         if len(state) != len(indices_list):
             raise ValueError(
@@ -431,9 +511,10 @@ class ZernikeMomentsBuilder(functions.Function):
             if get_jacobian:
                 n, l, m = zernike_idx
                 lin_idx = moms.linear_index(zernike_idx)
-                jac[lin_idx, idx] += 1 * num_type
+                jac[lin_idx, idx] += 1 * num_type  # pylint: disable=unsupported-assignment-operation
                 if zernike_idx.m > 0:
                     lin_idx = moms.linear_index((n, l, -m))
+                    # pylint: disable=unsupported-assignment-operation
                     jac[lin_idx, idx] += (-1)**m * (1 if num_type == 1. else -1j)
 
         if get_jacobian:
@@ -445,17 +526,19 @@ class ZernikeMomentsBuilder(functions.Function):
         # Skip all the ones that are masked
         if self._indices is None:
             # Lazily create the indices
-            self._indices = tuple(self._iter_indices(self._n_max))
+            self._indices = tuple(self._iter_indices(self._n_max, self._l_max))
 
         for zernike_index, num_type in self._indices:
             if self._mask is None or self._mask[zernike_index] is None:
                 yield zernike_index, num_type
 
     @classmethod
-    def _iter_indices(cls, n_max: int):
+    def _iter_indices(cls, n_max: int, l_max: int):
         # First deal with m = 0 as these are all real
+        indices = index_traits(n_max, l_max)
+
         for n in inclusive(0, n_max):
-            for l in inclusive(n):
+            for l in indices.iter_l(n):
                 if not even(n - l):
                     continue
 
@@ -464,7 +547,7 @@ class ZernikeMomentsBuilder(functions.Function):
         # Now deal with m > 0, these have both a real and imaginary part
         for num_type in (1., 1j):
             for n in inclusive(n_max):
-                for l in inclusive(n):
+                for l in indices.iter_l(n):
                     if not even(n - l):
                         continue
 
@@ -483,7 +566,7 @@ class ZernikeMomentsBuilder(functions.Function):
         def inverse(self) -> 'ZernikeMomentsBuilder':
             return self._forward
 
-        def evaluate(# pylint: disable=unused-argument
+        def evaluate(  # pylint: disable=unused-argument
                 self, state: ZernikeMoments, *, get_jacobian=False) -> np.ndarray:
             moms = []
             for indices, num_type in self._forward.iter_indices():
@@ -493,7 +576,7 @@ class ZernikeMomentsBuilder(functions.Function):
             return np.array(moms)
 
 
-class ZernikeMomentCalculator(base_moments.MomentsCalculator):
+class ZernikeMomentsCalculator(base_moments.MomentsCalculator):
     """Calculate Zernike moments.
 
     Takes as input geometric moments or any state vector that is a valid input to GeometricMomentsCalculator
@@ -502,31 +585,66 @@ class ZernikeMomentCalculator(base_moments.MomentsCalculator):
     supports_jacobian = True
     dtype = complex
     input_type = geometric.GeometricMoments, np.ndarray, functions.Features
+    _change_of_basis_polys = dict()  # Change of basis polynomials (geometric -> spherical harmonic)
 
-    def __init__(self, max_order: int):
+    @classmethod
+    def change_of_basis(cls, n: int, l: int, m: int) -> polynomials.HomogenousPolynomial:
+        try:
+            return cls._change_of_basis_polys[(n, l, m)]
+        except KeyError:
+            poly = sum_chi_nlm(n, l, m, polynomials.PolyBuilder())
+            cls._change_of_basis_polys[(n, l, m)] = poly
+
+            if m != 0 and (n, l, -m) not in cls._change_of_basis_polys:
+                # Cache the -m version as well
+                cls._change_of_basis_polys[(n, l, -m)] = (-1)**m * poly.conjugate()
+
+            return poly
+
+    def __init__(self, n_max: int, l_max: int = None, use_direct=False):
         super().__init__()
-        self._max_order = max_order
-        self._geometric_moments_calculator = geometric.GeometricMomentsCalculator(max_order)
+        self._n_max = n_max
+        self._l_max = l_max
+        self._geometric_moments_calculator = geometric.GeometricMomentsCalculator(n_max)
         self._jacobian = None
-        self._chi = None  # The omega_nl_m prefactors
+        self._use_direct = use_direct
 
     def evaluate(self,
                  state: functions.State,
                  *,
                  get_jacobian=False) -> Union[ZernikeMoments, Tuple[ZernikeMoments, np.ndarray]]:
+        if self._use_direct:
+            if not isinstance(state, functions.Features) or not all(
+                isinstance(feature, functions.WeightedDelta) for feature in state
+            ):
+                raise ValueError('Can only use direct method on a set of features that are delta functions')
+
+            return from_deltas_direct(
+                self._n_max,
+                positions=np.array([feature.pos for feature in state]),
+                weights=np.array([feature.weight for feature in state]),
+                lmax=self._l_max
+            )
+
         if isinstance(state, geometric.GeometricMoments):
             geom_moments = state
             geom_jac = np.eye(state.size)
         else:
+            # if not get_jacobian and isinstance(state, functions.Features) and \
+            #         all(isinstance(feature, functions.WeightedDelta) for feature in state):
+            #     positions = np.array([feature.pos for feature in state])
+            #     weights = np.array([feature.weight for feature in state])
+            #     return from_deltas_direct(self._n_max, positions, weights, lmax=self._l_max)
+
             geom_moments = self._geometric_moments_calculator(state, get_jacobian)
             if get_jacobian:
                 geom_moments, geom_jac = geom_moments
 
-        moments = ZernikeMoments(self._max_order, dtype=object if geom_moments.dtype == object else complex)
+        moments = ZernikeMoments(self._n_max, self._l_max, dtype=object if geom_moments.dtype == object else complex)
 
         # Get the moments themselves from polynomials of geometric moments
-        for (n, l, m), poly in self.chi.items():
-            moments[n, l, m] = poly.evaluate(geom_moments)
+        for (n, l, m) in moments.iter_indices(redundant=False):
+            moments[n, l, m] = omega_nl_m(n, l, m, geom_moments.array)
 
         if get_jacobian:
             jac = np.matmul(self._get_jacobian(), geom_jac)
@@ -537,47 +655,23 @@ class ZernikeMomentCalculator(base_moments.MomentsCalculator):
     def _get_jacobian(self):
         if self._jacobian is None:
             # Calculate the first time
-            self._jacobian = get_jacobian_wrt_geom_moments(self._max_order, redundant=True)
+            self._jacobian = get_jacobian_wrt_geom_moments(self._n_max, redundant=True)
 
         return self._jacobian
 
-    @property
-    def chi(self) -> Dict[base_moments.Index, base_moments.MomentsPolynomial]:
-        """Get the chi_nlm dictionary.  Indices are moment indices and values are polynomials"""
-        if self._chi is None:
-
-            class Tracker:
-                """Keeps track of the moments evaluated by omega_nl_m"""
-
-                def __init__(self, max_order):
-                    self._max_order = max_order
-
-                def __getitem__(self, item):
-                    mtx = np.zeros((self._max_order + 1, self._max_order + 1, self._max_order + 1), dtype=complex)
-                    mtx[item] = 1.
-                    return mtx
-
-            chi = collections.OrderedDict()
-            for n, l, m in iter_indices(self._max_order, redundant=False):
-                used = omega_nl_m(n, l, m, Tracker(self._max_order))
-                indexes = np.argwhere(used)
-                poly = base_moments.MomentsPolynomial()
-
-                for index in indexes:
-                    index = tuple(index)
-                    poly.append(base_moments.Product(used[index], (index,)))
-
-                chi[(n, l, m)] = poly
-
-            self._chi = chi
-
-        return self._chi
-
     def create_random(self, max_order: int = None) -> ZernikeMoments:
-        max_order = max_order or self._max_order
+        max_order = max_order or self._n_max
         return ZernikeMoments.rand(max_order)
 
 
+ZernikeMomentCalculator = ZernikeMomentsCalculator
+"""An alias left here for compatibility reasons"""
+
+
+def rand(max_order: int) -> ZernikeMoments:
+    return ZernikeMomentsCalculator(max_order).create_random()
+
+
 def get_jacobian_wrt_geom_moments(max_order: int, redundant=True):
     """Get the Jacobian of the Zernike moments wrt Geometric moments as input"""
 
@@ -604,59 +698,63 @@ def get_jacobian_wrt_geom_moments(max_order: int, redundant=True):
     num_geometric_moments = geometric.GeometricMoments.num_moments(max_order)
     jacobian = np.zeros((num_moments, num_geometric_moments), dtype=complex)
 
+    cache = {}  # Used for negative m coefficients
     for idx, (n, l, m) in enumerate(iter_indices(max_order, redundant=redundant)):
-        if m < 0:
-            # These are taken care of during the positive m iteration
-            continue
-
-        vec = omega_nl_m(n, l, m, tracker)
-        jacobian[idx, :] = vec
-
-        if redundant and m != 0:
-            minus_m = (-1)**m * vec.conjugate()
-            minus_m_idx = linear_index((n, l, -m))
-            jacobian[minus_m_idx, :] = minus_m
+        if redundant and m < 0:
+            jacobian[idx, :] = (-1)**(-m) * cache.pop((n, l, -m)).conjugate()
+        else:
+            vec = omega_nl_m(n, l, m, tracker)
+            jacobian[idx, :] = vec
+            if redundant and m > 0:
+                # Because of the iteration order (m=0 to l followed by m=-l to -1) the positives will always be
+                # calculated first, so cache these for when we get around to the positive m's
+                cache[(n, l, m)] = vec
 
     return jacobian
 
 
 @functools.lru_cache(maxsize=None)
 def factorial(n: int) -> int:
-    return scipy.special.factorial(n, exact=True)
+    return special.factorial(n, exact=True)
 
 
 @functools.lru_cache(maxsize=None)
 def binomial_coeff(n: int, k: int) -> int:
-    return scipy.special.comb(n, k, exact=True)
+    return special.comb(n, k, exact=False)
 
 
 @functools.lru_cache(maxsize=None)
 def c_l_m(l: int, m: int) -> float:
     """Calculate the normalisation factor"""
     if m < 0:
-        # make use of symmetry c_l^m == c_l^-m
+        # make use of symmetry c_l^m == c_l^-m (no conjugate because these are real)
         return c_l_m(l, -m)
 
-    return ((2 * l + 1) * factorial(l + m) * factorial(l - m))**0.5 / factorial(l)
+    return math.sqrt((2 * l + 1) * factorial(l + m) * factorial(l - m)) / factorial(l)
 
 
 @functools.lru_cache(maxsize=None)
 def q_kl_nu(k: int, l: int, nu: int) -> float:
     """Calculate the Zernike geometric moment conversion factors"""
     return (-1) ** k / 2 ** (2 * k) * \
-           ((2 * l + 4 * k + 3) / 3) ** 0.5 * \
+           np.sqrt((2 * l + 4 * k + 3) / 3) * \
            binomial_coeff(2 * k, k) * \
            (-1) ** nu * \
            (binomial_coeff(k, nu) * binomial_coeff(2 * (k + l + nu) + 1, 2 * k)) / \
            binomial_coeff(k + l + nu, k)
 
 
-def sum_chi_nlm(n: int, l: int, m: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
-    """Calculate the Zernike geometric moment conversion factors"""
+def sum_chi_nlm(n: int, l: int, m: int, geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
+    """This call performs the change of basis transformation from a polynomial basis to Zernike"""
+    assert_valid(n, l, m)
+    if m < 0:
+        # Symmetry relation
+        return (-1)**(-m) * sum_chi_nlm(n, l, -m, geom_moments).conjugate()
+
     return c_l_m(l, m) * 2**(-m) * sum1(n=n, l=l, m=m, geom_moments=geom_moments)
 
 
-def sum1(n: int, l: int, m: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
+def sum1(n: int, l: int, m: int, geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
     """Calculate the Zernike geometric moment conversion factors"""
     k = int(math.floor((n - l) / 2))  # n - l is always even
     total = 0.
@@ -666,7 +764,7 @@ def sum1(n: int, l: int, m: int, geom_moments: np.ndarray) -> Union[complex, np.
     return total
 
 
-def sum2(n: int, l: int, m: int, nu: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
+def sum2(n: int, l: int, m: int, nu: int, geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
     total = 0.
     for alpha in inclusive(nu):
         total += binomial_coeff(nu, alpha) * \
@@ -674,7 +772,8 @@ def sum2(n: int, l: int, m: int, nu: int, geom_moments: np.ndarray) -> Union[com
     return total
 
 
-def sum3(n: int, l: int, m: int, nu: int, alpha: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
+def sum3(n: int, l: int, m: int, nu: int, alpha: int,
+         geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
     total = 0.
     for beta in inclusive(nu - alpha):
         total += binomial_coeff(nu - alpha, beta) * \
@@ -682,7 +781,8 @@ def sum3(n: int, l: int, m: int, nu: int, alpha: int, geom_moments: np.ndarray)
     return total
 
 
-def sum4(n: int, l: int, m: int, nu: int, alpha: int, beta: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
+def sum4(n: int, l: int, m: int, nu: int, alpha: int, beta: int,
+         geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
     total = 0.
     for u in inclusive(m):
         total += (-1) ** (m - u) * binomial_coeff(m, u) * 1j ** u * \
@@ -692,7 +792,7 @@ def sum4(n: int, l: int, m: int, nu: int, alpha: int, beta: int, geom_moments: n
 
 
 def sum5(n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int,
-         geom_moments: np.ndarray) -> Union[complex, np.array]:
+         geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
     total = 0.
     for mu in inclusive(int(math.floor((l - m) / 2.))):
         total += (-1) ** mu * 2 ** (-2 * mu) * binomial_coeff(l, mu) * \
@@ -702,8 +802,9 @@ def sum5(n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int,
     return total
 
 
-def sum6(n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int, mu: int,
-         geom_moments: np.ndarray) -> Union[complex, np.array]:
+def sum6(
+    n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int, mu: int, geom_moments: base_moments.MomentsProto
+) -> Union[complex, np.array]:
     total = 0.
     for v in inclusive(mu):
         r = 2 * (v + alpha) + u
@@ -729,7 +830,7 @@ def assert_valid(n: int, l: int, m: int):
         raise ValueError('n - l must be even')
 
 
-def omega_nl_m(n: int, l: int, m: int, geom_moments: np.array) -> complex:
+def omega_nl_m(n: int, l: int, m: int, geom_moments: base_moments.MomentsProto) -> complex:
     """Given a set of geometric moments this function will compute the corresponding Zernike moments
     """
     assert_valid(n, l, m)
@@ -737,24 +838,23 @@ def omega_nl_m(n: int, l: int, m: int, geom_moments: np.array) -> complex:
         # Symmetry relation
         return (-1)**(-m) * omega_nl_m(n, l, -m, geom_moments).conjugate()
 
-    return 3. / (4. * np.pi) * sum_chi_nlm(n, l, m, geom_moments).conjugate()
+    return NORMALISTAION * ZernikeMomentsCalculator.change_of_basis(n, l, m)(geom_moments).conjugate()
 
 
-def iter_indices(max_order: int = None, redundant=False) -> Iterator[ZernikeIndex]:
+def iter_indices(n_max: int = None, l_max: int = None, *, redundant=False) -> Iterator[ZernikeIndex]:
     """Iterate over Zernike function indices in lexicographic order.
 
     If redundant is True then all valid indices will be generated including those where there is a
     symmetry relation e.g.: o_22^-2 = (-1)^2 o22^2.
 
-    :param max_order: order to provide indices up to, if None will generate all indices
+    :param n_max: order to provide indices up to, if None will generate all indices
+    :param l_max: the maximum angular frequency to go up to
     :param redundant: if True include redundant indices
     """
-    upper = itertools.count() if max_order is None else inclusive(max_order)
+    upper = itertools.count() if n_max is None else inclusive(n_max)
     for n in upper:
-        for l in inclusive(n):
-            if not even(n - l):
-                continue
-
+        indices = index_traits(n, l_max if l_max is None else min(n, l_max))
+        for l in indices.iter_l(n):
             for m in inclusive(0, l):
                 yield ZernikeIndex(n, l, m)
 
@@ -766,6 +866,7 @@ def iter_indices(max_order: int = None, redundant=False) -> Iterator[ZernikeInde
 
 
 def iter_indices_extended(key: Tuple[int, Tuple], max_order: int, redundant=False) -> Iterator[ZernikeIndex]:
+    # pylint: disable=too-many-branches
     if isinstance(key, int):
         # Assume the caller is treating us like a vector
         n, l, m = triple_index(key)
@@ -811,13 +912,14 @@ def triple_index(linear: int, redundant=True) -> ZernikeIndex:
     entry = None
     for entry, _ in zip(iter_indices(redundant=redundant), range(linear + 1)):
         pass
-    return entry
+    return entry  # pylint: disable=inconsistent-return-statements
 
 
 @functools.lru_cache(maxsize=256)
 def linear_index(index: base_moments.Index, redundant=True) -> int:
     """Given a triple of zernike function indices this will return a lexicographically ordered
     integer"""
+    # pylint: disable=inconsistent-return-statements
     if not redundant and index[2] < 0:
         raise ValueError('m value cannot be negative')
 
@@ -825,8 +927,7 @@ def linear_index(index: base_moments.Index, redundant=True) -> int:
         if triple == index:
             return linear
 
-    # pylint: disable=inconsistent-return-statements
-    assert False, 'Should never reach here'
+    raise RuntimeError('Should never reach here!')
 
 
 def _domain_check(positions: np.array):
@@ -841,3 +942,54 @@ def _domain_check(positions: np.array):
                 'Delta function %i is outside of the domain of Zernike basis functions '
                 'with are defined within the unit ball', idx
             )
+
+
+def index_traits(n, l) -> sph.IndexTraits:
+    return sph.IndexTraits(n, l, l_le_n=True, n_minus_l_even=True)
+
+
+def zernike_poly(n: int, l: int, m: int, pt, condon_shortley=True):
+    """Evaluate the 3D Zernike polynomial at the given spherical coordinates"""
+    rho = pt[0]
+    theta = pt[1]  # need to be [0, pi]
+    # I'm not sure why but there seems to be a pi/2 phase difference when using sph_harm, so just include it here
+    phi = pt[2] + np.pi / 2
+
+    theta = np.where(theta < 0, np.pi + theta, theta)
+    phi = np.where(phi < 0, np.pi + phi, phi)
+
+    radial = r_nl(n, l, rho)
+    # Note that scipy uses the opposite of the physics convention, in scipy
+    #   theta = azimuth (longitudinal) coord
+    #   phi = polar (colatitudinal) coord
+    # while we use the opposite convention
+    angular = special.sph_harm(m, l, phi, theta)
+    if not condon_shortley:
+        # scipy sph_harm incorporates the Condon-Shortley phase so if we don't want it we have to
+        # cancel it out here
+        angular /= (-1)**m
+
+    return radial * angular
+
+
+def r_nl(n: int, l: int, rho) -> numbers.Number:
+    """Evaluate a Zernike radial polynomial at the given value of rho"""
+    if not even(n - l):
+        return 0.
+    #
+    # D = 3.
+    # k = (n - l) / 2.
+    #
+    # normalisation = np.sqrt(2 * n + D)
+    # total = np.sum([((-1)**s) * binomial_coeff(k, s) * binomial_coeff(n - s - 1. + D / 2., k) * rho**(n - 2 * s)
+    #                 for s in inclusive(int(k))],
+    #                axis=0)
+    #
+    # # Alternative method
+    # # total2 = (-1)**k * binomial_coeff((n + l + D) / 2 - 1, k) * rho**l * \
+    # #          special.hyp2f1(-k, (n + l + D) / 2., l + D / 2., rho**2)
+    #
+    # return normalisation * total
+    # #
+    k = int((n - l) / 2)
+    return rho**l * sum(q_kl_nu(k, l, nu) * rho**(2 * nu) for nu in inclusive(k))
